# -*- coding: utf-8 -*-

from dataclasses import dataclass
import copy
from enum import Enum
import glob
from itertools import combinations
import logging
import os
import re
import string
from typing import Any, List, Mapping, Union

from gordias.config import get_configuration
import iris
import xdg.BaseDirectory
import yaml
from importlib.metadata import entry_points

from .index import Index
from .period import PeriodSpecification


@dataclass(eq=True, frozen=True)
class Distribution:
    name: str
    version: str


@dataclass
class CellMethod:
    name: str
    method: str


def format_var_name(var_name, parameters):
    def format_value(value):
        if value < 0:
            return f"m{-value}"
        else:
            return f"{value}"

    parsed_var_name = list(string.Formatter().parse(var_name))
    items = {
        ft[1]: format_value(parameters[ft[1]])
        for ft in parsed_var_name
        if ft[1] is not None
    }
    return var_name.format(**items)


@dataclass
class OutputVariable:
    var_name: str
    standard_name: str
    proposed_standard_name: str
    long_name: str
    units: str
    cell_methods: List[CellMethod]

    @property
    def drs(self):
        return {"var_name": self.var_name}

    def instantiate(self, parameters):
        return OutputVariable(
            format_var_name(self.var_name, parameters),
            self.standard_name,
            self.proposed_standard_name,
            self.long_name.format(**parameters),
            self.units,
            self.cell_methods,
        )


@dataclass
class InputVariable:
    var_name: str
    standard_name: str
    cell_methods: List[CellMethod]
    aliases: List[str]

    def instantiate(self, parameters):
        return InputVariable(
            format_var_name(self.var_name, parameters),
            self.standard_name,
            self.cell_methods,
            self.aliases,
        )


def build_variable(name, variable, path):
    cell_methods = [CellMethod(*cm.popitem()) for cm in variable.pop("cell_methods")]
    return InputVariable(
        name, variable["standard_name"], cell_methods, variable["aliases"]
    )


class ParameterKind(Enum):
    FLAG = "flag"
    OPERATOR = "operator"
    QUANTITY = "quantity"
    REDUCER = "reducer"
    TIME_RANGE = "time_range"


PARAMETER_KINDS = {}


@dataclass
class ParameterFlag:
    name: str
    flag: bool
    kind: ParameterKind = ParameterKind.FLAG

    @property
    def parameter(self):
        return self.flag

    def instantiate(self, parameters):
        return ParameterFlag(ParameterKind.FLAG, self.flag)


PARAMETER_KINDS["flag"] = ParameterFlag


@dataclass
class ParameterOperator:
    name: str
    operator: str
    kind: ParameterKind = ParameterKind.OPERATOR

    @property
    def parameter(self):
        return self.operator

    def instantiate(self, parameters):
        return ParameterOperator(ParameterKind.OPERATOR, self.operator)


PARAMETER_KINDS["operator"] = ParameterOperator


@dataclass
class ParameterQuantity:
    var_name: str
    standard_name: str
    data: Any
    units: str
    long_name: str = None
    proposed_standard_name: str = None
    kind: ParameterKind = ParameterKind.QUANTITY

    @property
    def parameter(self):
        return iris.coords.AuxCoord(
            self.data,
            self.standard_name,
            self.long_name,
            self.var_name,
            units=self.units,
        )

    def instantiate(self, parameters):
        data = self.data
        if isinstance(data, dict) and len(data) == 1:
            key, value = data.popitem()
            data = parameters[key]
        elif isinstance(data, str):
            data = int(data.format(**parameters))
        ln = self.long_name
        if ln is not None:
            ln = ln.format(**parameters)
        param = ParameterQuantity(
            format_var_name(self.var_name, parameters),
            self.standard_name,
            data,
            self.units,
            ln,
        )
        return param


PARAMETER_KINDS["quantity"] = ParameterQuantity


@dataclass
class ParameterReducer:
    name: str
    reducer: str
    kind: ParameterKind = ParameterKind.REDUCER

    @property
    def parameter(self):
        return self.reducer

    def instantiate(self, parameters):
        return ParameterReducer(ParameterKind.REDUCER, self.reducer)


PARAMETER_KINDS["reducer"] = ParameterReducer


@dataclass
class ParameterTimeRange:
    name: str
    data: str
    kind: ParameterKind = ParameterKind.TIME_RANGE

    @property
    def parameter(self):
        return self.data

    def instantiate(self, parameters):
        return ParameterTimeRange(ParameterKind.TIME_RANGE, self.data)


PARAMETER_KINDS["time_range"] = ParameterTimeRange


Parameter = Union[
    ParameterQuantity, ParameterOperator, ParameterReducer, ParameterTimeRange
]


@dataclass
class IndexFunction:
    name: str
    parameters: Mapping[str, Parameter]

    def instantiate(self, parameters):
        return IndexFunction(
            self.name,
            {
                name: param.instantiate(parameters)
                for name, param in self.parameters.items()
            },
        )


@dataclass
class IndexDefinition:
    reference: str
    default_period: str
    output: OutputVariable
    input: Mapping[str, InputVariable]
    index_function: IndexFunction
    source: str
    distribution: Distribution

    def instantiate(self, parameters):
        idx = IndexDefinition(
            self.reference,
            self.default_period,
            self.output.instantiate(parameters),
            {key: iv.instantiate(parameters) for key, iv in self.input.items()},
            self.index_function.instantiate(parameters),
            self.source,
            self.distribution,
        )
        return idx


def build_parameter(name, metadata):
    return PARAMETER_KINDS[metadata["kind"]](name, **metadata)


def build_index(metadata, variables, source=None, distribution=None):
    if (cell_methods := metadata["output"]["cell_methods"]) is None:  # noqa
        cms = []
    else:
        cms = [CellMethod(*cm.popitem()) for cm in cell_methods]
    output = OutputVariable(
        metadata["output"]["var_name"],
        metadata["output"]["standard_name"],
        metadata["output"].get("proposed_standard_name", None),
        metadata["output"]["long_name"],
        metadata["output"]["units"],
        cms,
    )
    if isinstance(metadata["input"], str):
        input_metadata = {"data": metadata["input"]}
    else:
        input_metadata = metadata["input"]
    input = {key: variables[name] for key, name in input_metadata.items()}
    params = metadata["index_function"]["parameters"]
    if params is None:
        parameters = {}
    else:
        parameters = {name: build_parameter(name, params[name]) for name in params}
    index_function = IndexFunction(metadata["index_function"]["name"], parameters)
    idx = IndexDefinition(
        metadata["reference"],
        metadata["default_period"],
        output,
        input,
        index_function,
        source,
        distribution,
    )
    return idx


def get_signature_candidates(signature_parts):
    m_indices = [i for i, p in enumerate(signature_parts) if p.endswith("m")]
    combs = sum(
        [list(combinations(m_indices, i)) for i in range(len(m_indices) + 1)], []
    )
    candidates = []
    for combination in combs:
        parts = copy.copy(signature_parts)
        for idx in combination:
            parts[idx] = parts[idx][:-1]
        candidates.append(tuple(parts))
    return zip(combs, candidates)


class IndexCatalog:
    def __init__(self, indices):
        self.indices = indices
        self._build_template_index()

    def get_list(self):
        return self.indices.keys()

    def _build_template_index(self):
        expr = re.compile(r"{([^}]+)}")
        template_index = {}
        for index in self.indices.keys():
            split = expr.split(index)
            if len(split) == 1:
                continue
            signature = tuple(split[::2])
            parameter_names = split[1::2]
            template_index[signature] = (index, parameter_names)
        self.template_index = template_index

    def get_index_definition(self, index):
        try:
            return self.indices[index]
        except KeyError:
            index_expr = re.compile(r"(\d+)")
            split = index_expr.split(index)
            if len(split) == 1:
                raise
            signature_parts = split[::2]
            candidates = get_signature_candidates(signature_parts)
            matching_signatures = [
                candidate
                for candidate in candidates
                if candidate[1] in self.template_index
            ]
            if len(matching_signatures) == 0:
                raise
            elif len(matching_signatures) > 1:
                raise RuntimeError("More than one matching signature found")
            combination, signature = matching_signatures[0]
            template_name, parameter_names = self.template_index[signature]
            parameter_values = split[1::2]
            for i in combination:
                parameter_values[i] = "-" + parameter_values[i]
            parameter_dict = {
                name: int(value)
                for (name, value) in zip(parameter_names, parameter_values)
            }
            template = self.indices[template_name]
            index_definition = template.instantiate(parameter_dict)
            return index_definition

    def prepare_indices(
        self,
        requested_indices,
        period=None,
        reference_period=None,
        computational_period=None,
    ):
        def select_period(period_string):
            period_parts = period_string.split("[")
            period_type = period_parts[0]
            if len(period_parts) == 2 and period_parts[1][-1] == "]":
                if period_type == "annual":
                    period_constraints = period_parts[1][:-1]
                else:
                    raise ValueError(
                        f"Period constraints <{period_parts[1][:-1]}> can only be used "
                        "with `annual` period type. The following period type was used "
                        f"<{period_type}>."
                    )
            else:
                period_constraints = None
            if computational_period is not None:
                computational_constraint = computational_period
            else:
                computational_constraint = None
            return PeriodSpecification(
                period_type, (period_constraints, computational_constraint)
            )

        indices = []
        for index_name in requested_indices:
            definition = self.get_index_definition(index_name)
            period_spec = select_period(
                definition.default_period if period is None else period
            )
            if reference_period is not None:
                if "reference_period" not in definition.index_function.parameters:
                    logging.warning(
                        f"Trying to apply a reference period to index <{index_name}>, "
                        f"but <{index_name}> doesn't use it."
                    )
                else:
                    definition.index_function.parameters[
                        "reference_period"
                    ].data = reference_period
            try:
                index_function = build_index_function(definition.index_function)
            except TypeError:
                logging.error(
                    f"Could not build index function for index "
                    f"{index_name} from definition {definition}"
                )
                raise
            logging.info(
                f"Trying to build index <{index_name}> "
                f"with period <{period_spec.type}"
                f"{p if (p := period_spec.parameters) is not None else ''}> "
                f"from definition in <{definition.distribution.name}-"
                f"{definition.distribution.version}> ({definition.source})."
            )
            index = Index(index_function, definition, period_spec)
            indices.append(index)
        return indices


def build_index_function(spec):
    name = spec.name
    candidates = list(entry_points(group="climix.index_functions", name=name))
    if len(candidates) == 0:
        raise ValueError(f"No implementation found for index_function <{name}>")
    elif len(candidates) > 1:
        distributions = [candidate.dist for candidate in candidates]
        raise ValueError(
            f"Found several implementations for index_function <{name}>. "
            f"Please make sure only one is installed at any time. "
            f"The implementations come from the distributions {distributions}"
        )
    candidate = candidates[0]
    logging.info(
        f"Found implementation for index_function <{name}> "
        f"from distribution <{candidate.dist}>"
    )
    index_function_factory = candidates[0].load()
    parameters = {name: param.parameter for name, param in spec.parameters.items()}
    index_function = index_function_factory(**parameters)
    return index_function


def build_distribution(metadata):
    dist_metadata = metadata.get("distribution", {})
    return Distribution(dist_metadata.get("name"), dist_metadata.get("version"))


def find_metadata_files_in_dir(directory):
    if os.path.isdir(directory):
        return glob.glob(os.path.join(directory, "*.yml"))
    return []


def find_metadata_files(metadata_files):
    directories = [
        os.path.join(os.path.dirname(__file__), "etc"),
        "/etc/climix",
    ] + list(xdg.BaseDirectory.load_config_paths("climix"))[::-1]
    for d in directories:
        logging.info(f"Looking for metadata in directory {d}")
    files = sum(
        [find_metadata_files_in_dir(directory) for directory in directories], []
    )
    if metadata_files is not None:
        for f in metadata_files:
            logging.info(f"Adding metadata from file: {f}")
            files.append(f)
    return files


def load_metadata(metadata_files=None, activate_config=False):
    variables = {}
    indices = {}
    configuration = None
    variable_metadata = []
    index_metadata = []
    configuration_metadata = []
    for path in find_metadata_files(metadata_files):
        logging.info(f"Reading metadata definitions from file {path}")
        with open(path) as md_file:
            metadata = yaml.safe_load(md_file)
        dist = build_distribution(metadata)
        index_metadata.append((metadata.get("indices", {}), path, dist))
        variable_metadata.append((metadata.get("variables", {}), path, dist))
        configuration_metadata.append((metadata.get("config", {}), path))
    for var_metadata, path, dist in variable_metadata:
        for name, var in var_metadata.items():
            if name in variables:
                logging.info(
                    f"Replacing variable definition for <{name}> with "
                    f"definition from distribution <{dist.name}-{dist.version}> "
                    f"({path})."
                )
            variables[name] = build_variable(name, var, path)
    for idx_metadata, path, dist in index_metadata:
        for name, idx_meta in idx_metadata.items():
            if name in indices:
                logging.info(
                    f"Replacing index definition for <{name}> with "
                    f"definition from distribution <{dist.name}-{dist.version}> "
                    f"({path})."
                )
            try:
                indices[name] = build_index(idx_meta, variables, path, dist)
            except (KeyError, TypeError):
                logging.error("Metadata error for index {} from {}.".format(name, path))
                raise
    if activate_config:
        current_config = None
        for config_metadata, path in configuration_metadata:
            if config_metadata:
                current_config = config_metadata
                current_config_path = path
        if current_config is not None:
            logging.info(
                "Building configuration with definition from file "
                f"<{current_config_path}>."
            )
            try:
                configuration = get_configuration(current_config)
            except (KeyError, TypeError):
                logging.error(
                    f"Metadata error for climix configuration <{current_config}> from "
                    f"<{current_config_path}>."
                )
                raise
    return IndexCatalog(indices), configuration



# -*- coding: utf-8 -*-

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("climix")
except PackageNotFoundError:
    # package is not installed
    pass

__all__ = [
    "__version__",
]



import pytest
from dask.distributed import LocalCluster
from iris.cube import Cube
from iris.coords import AuxCoord
import numpy as np
from cf_units import Unit
from iris.coords import DimCoord


def add_dim_coord(cube, time):
    """Add time dimension coordinates to cube."""
    shape = cube.data.shape
    assert len(shape) == 3
    time_coord = DimCoord(
        time,
        var_name="time",
        standard_name="time",
        units="days since 1949-12-01 00:00:00",
    )
    cube.add_dim_coord(time_coord, 0)


@pytest.fixture(scope="module")
def f_dask_cluster():
    cluster = LocalCluster(n_workers=2, threads_per_worker=2, dashboard_address=None)
    yield cluster
    cluster.close()


@pytest.fixture
def f_time_cube_tas(request):
    test_cube = Cube(
        data=request.param["data"].astype(np.float32),
        standard_name="air_temperature",
        var_name="tas",
        units=request.param["units"],
    )
    if "time" in request.param:
        add_dim_coord(test_cube, request.param["time"])
    return test_cube


@pytest.fixture
def f_cube_tas(request):
    test_cube = Cube(
        data=request.param["data"].astype(np.float32),
        standard_name="air_temperature",
        var_name="tas",
        units=request.param["units"],
    )
    return test_cube


@pytest.fixture
def f_cube_tasmax(request):
    test_cube = Cube(
        data=request.param["data"].astype(np.float32),
        standard_name="air_temperature",
        var_name="tasmax",
        units=request.param["units"],
    )
    return test_cube


@pytest.fixture
def f_cube_tasmin(request):
    test_cube = Cube(
        data=request.param["data"].astype(np.float32),
        standard_name="air_temperature",
        var_name="tasmin",
        units=request.param["units"],
    )
    return test_cube


@pytest.fixture
def f_cube_pr(request):
    test_cube = Cube(
        data=request.param["data"].astype(np.float32),
        standard_name=request.param["standard_name"],
        var_name="pr",
        units=request.param["units"],
    )
    return test_cube


@pytest.fixture
def f_first_threshold(request):
    long_name = "first_threshold"
    if "long_name" in request.param:
        long_name = request.param["long_name"]
    aux_coord = AuxCoord(
        np.array([request.param["data"]]),
        standard_name=request.param["standard_name"],
        units=Unit(request.param["units"]),
        var_name="first_threshold",
        long_name=long_name,
    )
    return aux_coord


@pytest.fixture
def f_second_threshold(request):
    long_name = "second_threshold"
    if "long_name" in request.param:
        long_name = request.param["long_name"]
    aux_coord = AuxCoord(
        np.array([request.param["data"]]),
        standard_name=request.param["standard_name"],
        units=Unit(request.param["units"]),
        var_name="second_threshold",
        long_name=long_name,
    )
    return aux_coord


@pytest.fixture
def f_third_threshold(request):
    long_name = "third_threshold"
    if "long_name" in request.param:
        long_name = request.param["long_name"]
    aux_coord = AuxCoord(
        np.array([request.param["data"]]),
        standard_name=request.param["standard_name"],
        units=Unit(request.param["units"]),
        var_name="third_threshold",
        long_name=long_name,
    )
    return aux_coord


@pytest.fixture
def f_percentile(request):
    aux_coord = AuxCoord(
        np.array([request.param["data"]]),
        units=Unit(request.param["units"]),
    )
    return aux_coord


@pytest.fixture
def f_window_size(request):
    aux_coord = AuxCoord(
        np.array([request.param["data"]]),
        units=Unit(request.param["units"]),
    )
    return aux_coord


@pytest.fixture
def f_first_duration(request):
    aux_coord = AuxCoord(
        np.array([request.param["data"]]),
        units=Unit(request.param["units"]),
    )
    return aux_coord


@pytest.fixture
def f_second_duration(request):
    aux_coord = AuxCoord(
        np.array([request.param["data"]]),
        units=Unit(request.param["units"]),
    )
    return aux_coord


@pytest.fixture
def f_first_delay(request):
    aux_coord = AuxCoord(
        np.array([request.param["data"]]),
        units=Unit(request.param["units"]),
    )
    return aux_coord


@pytest.fixture
def f_second_delay(request):
    aux_coord = AuxCoord(
        np.array([request.param["data"]]),
        units=Unit(request.param["units"]),
    )
    return aux_coord



# -*- coding: utf-8 -*-

from .change_pr_units import is_precipitation, change_pr_units


def change_units(cube_or_coord, new_units, new_standard_name=None):
    if is_precipitation(new_standard_name):
        change_pr_units(cube_or_coord, new_units, new_standard_name)
    else:
        cube_or_coord.convert_units(new_units)


__all__ = [
    change_units,
]



import logging

from cf_units import Unit
import dask.array as da
from gordias.dask_setup import progress
import numpy as np

from ..dask_take_along_axis import dask_take_along_axis
from .support import (
    IndexFunction,
    normalize_axis,
    ReferencePeriod,
    NUMPY_OPERATORS,
    DASK_OPERATORS,
)
from ..util import change_units


def calc_doy_indices(day_of_year):
    max_doy = day_of_year.max()
    for doy in range(1, max_doy + 1):
        exact_inds = np.nonzero(day_of_year == doy)[0]
        inds = np.stack(
            [exact_inds + i for i in range(-2, 3)],
            axis=-1,
        ).ravel()
        yield doy, inds


def convert_to_desired_output(self, data, masked_data, axis, output_unit):
    self.units = Unit(output_unit)
    if output_unit == "mm":
        return np.sum(masked_data, axis=axis)
    if output_unit == "1":
        return np.count_nonzero(masked_data, axis=axis)
    if output_unit == "%":
        sum_masked_data = np.sum(masked_data, axis=axis)
        return 100.0 * (sum_masked_data / np.sum(data, axis=axis))
    raise ValueError(
        f"Output unit <{output_unit}> is not supported for index "
        "function <count_thresholded_percentile_occurrences>."
    )


def hyndman_fan_params(prob, n):
    r"""
    Compute coefficient and index for quantile calculation.

    Following Hyndman and Fan (1996), the quantile for probability :math:`p`,
    :math:`Q(p)`, given the order statistics of a sample with size :math:`n`,
    :math:`X_{(1)} < \dots < X_{(n)}` can be calculated as a linear combination of two
    neighbouring order statistics according to

    .. math:: Q(p) = (1-\gamma)X_{(j)} + \gamma X_{(j+1)},

    where :math:`j = \lfloor pn + m \rfloor` and :math:`\gamma = pn + m - j`, i.e.
    :math:`j` and :math:`\gamma` are the integer and the fractional part of
    :math:`pn + m`, respectively, with :math:`m \in \mathbb{R}` some number such that

    .. math:: \frac{j - m}{n} \le p \le \frac{j - m + 1}{n}

    holds. Choosing :math:`m` now allows one to define different ways to calculate the
    quantiles. In particular, Hyndman and Fan go on to define a class of such approaches
    with

    .. math:: m = \alpha + p(1 - \alpha - \beta)

    characterized by the choice of the constants :math:`\alpha` and :math:`\beta`.
    For consistency with previous methods, here we follow there Definition 8 with
    :math:`\alpha = \frac{1}{3}`, :math:`\beta = \frac{1}{3}`.

    Defining :math:`\delta(\alpha, \beta) = pn + m = p(n + 1 - \alpha - \beta) +
    \alpha`, we find :math:`\delta(\frac{1}{3}, \frac{1}{3}) = p(n + \frac{1}{3}) +
    \frac{1}{3}`, which allows us to calculate :math:`j` and :math:`\gamma`.
    """
    n_array = da.asanyarray(n)
    delta = prob * (n_array + 1.0 / 3.0) + 1.0 / 3.0
    j = delta.astype(np.int64)
    gamma = delta - j
    return (j, gamma.astype(np.float64))


def hyndman_fan_to_topk(n, j, window_size):
    r"""
    Calculate topk argument for Hyndman and Fan order statistics index.

    We need order statistics :math:`j` and :math:`j + 1` to calculate the quantile in
    question. In a parallel setting, it is more efficient to avoid a full sorting of the
    input data and instead to only extract the :math:`j` largest or :math:`j + 1`
    smallest data points, particularly if :math:`j` is small or close to :math:`n` as is
    the case for the often sought extreme quantiles.
    If :math:`j > \frac{n}{2}` we want to calculate the :math:`n - j` largest points,
    else we want to calculate the :math:`j + 1` smallest points, indicated by the
    negative `k` (cf :func:`da.topk`).

    Calculating the quantile on input data that contains masked values, the sample size
    :math:`n` should specify the number of unmasked values. To account for the number of
    masked values we can use :math:`n_masked_values`. When calculating the window for
    the smallest data points, i.e, :math:`k_ext`, we need to include the number of
    masked values. This method assumes that the masked input data is filled with a
    relatively large negative value in relation to the input data. If the input data
    does not contain any masked values :math:`n_masked_values` should be set to zero.
    """
    if da.array((j > n / 2)).any():
        k = n - j
        k_ext = np.max(k + window_size)
    else:
        k = -(j + 1)
        k_ext = np.min(k - window_size)
    return (da.asanyarray(k).astype(np.int64), k_ext.astype(np.int64))


class BootstrapQuantiles:
    def __init__(self, data, prob, first_year, window_size, client, axis):
        data_mask = da.ma.getmaskarray(data)
        if data_mask.any():
            logging.warning(
                "Masked input data is not supported for index function "
                "<count_percentile_occurrences>."
            )
        self.axis = axis
        self.first_year = first_year
        n = data.shape[axis]
        self.j, self.gamma = hyndman_fan_params(prob, n)
        self.k, k_ext = hyndman_fan_to_topk(n, self.j, window_size)
        order_indices = client.persist(da.argtopk(data, k_ext, axis=self.axis))
        progress(order_indices)
        self.order_statistics = client.persist(
            dask_take_along_axis(data, order_indices, axis=self.axis).rechunk()
        )
        progress(self.order_statistics)
        self.years = client.persist(
            first_year + da.floor(order_indices / window_size).rechunk()
        )
        progress(self.years)

    def quantiles(self, ignore_year=None, duplicate_year=None):
        abs_k = abs(self.k)
        if ignore_year is None and duplicate_year is None:
            if self.k.ndim == 0:
                if abs_k < 1:
                    quantiles = self.order_statistics[..., abs_k]
                else:
                    if self.k >= 0:
                        ind_j = abs_k
                        ind_j_plus_one = abs_k - 1
                    else:
                        ind_j = abs_k - 1
                        ind_j_plus_one = abs_k
                    quantiles = (1.0 - self.gamma) * self.order_statistics[
                        ..., ind_j
                    ] + self.gamma * self.order_statistics[..., ind_j_plus_one]
            else:
                if (abs_k < 1).any():
                    ind_j = abs_k
                    ind_j_plus_one = da.where(abs_k > 0, abs_k + 1, abs_k)
                else:
                    if (self.k > 0).any():
                        ind_j = abs_k
                        ind_j_plus_one = abs_k - 1
                    else:
                        ind_j = abs_k - 2.0
                        ind_j_plus_one = abs_k - 1.0
                qi = dask_take_along_axis(
                    self.order_statistics,
                    da.stack([ind_j, ind_j_plus_one], axis=self.axis),
                    axis=self.axis,
                )
                quantiles = (1.0 - self.gamma) * qi[..., 0] + self.gamma * qi[..., 1]
        else:
            offset = da.sum(self.years[..., :abs_k] == ignore_year, axis=self.axis)
            offset -= da.sum(self.years[..., :abs_k] == duplicate_year, axis=self.axis)
            offset += abs_k
            if self.k >= 0:
                ind_j = offset
                ind_j_plus_one = da.where(offset >= 1, offset - 1, offset)
            else:
                ind_j = da.where(offset >= 1, offset - 1, offset)
                ind_j_plus_one = offset
            qi = dask_take_along_axis(
                self.order_statistics,
                da.stack([ind_j, ind_j_plus_one], axis=self.axis),
                axis=self.axis,
            )
            quantiles = (1.0 - self.gamma) * qi[..., 0] + self.gamma * qi[..., 1]
        return quantiles.astype(np.float32)


class BootstrapQuantilesThreshold(BootstrapQuantiles):
    def __init__(
        self, data, threshold, data_condition, prob, window_size, client, axis
    ):
        self.axis = axis
        cond = data_condition(data, threshold)
        n_effective = client.persist(da.count_nonzero(cond, axis=self.axis))
        progress(n_effective)
        self.j, self.gamma = hyndman_fan_params(prob, n_effective)
        self.k, k_ext = hyndman_fan_to_topk(n_effective, self.j, window_size)
        if k_ext < 0:
            fill_value = np.finfo(data.dtype).max
        else:
            fill_value = np.finfo(data.dtype).min
        input_data = da.where(cond, data, fill_value)
        del cond
        order_indices = client.persist(da.argtopk(input_data, k_ext, axis=self.axis))
        progress(order_indices)
        del input_data
        self.order_statistics = client.persist(
            dask_take_along_axis(data, order_indices, axis=self.axis).rechunk()
        )
        progress(self.order_statistics)

    def quantiles(self):
        return super().quantiles()


def build_indices(time, max_doy, no_years, window_size):
    """
    Build indices

    Given a linear time coordinate, build an index array `idx` of shape
    `(max_doy, no_years, window_size)` such that `idx[doy, yr]` contains the
    indices of the `window_size` days in the time coordinate that should
    contribute to the day of year `doy` for the year `yr`. If `doy` is smaller
    than `window_size`, this will include days from the year `yr -
    1`. Conversely, if `doy` is larger than `max_doy - window_size`, it will
    include days from `yr + 1`.
    """
    window_width = window_size // 2
    first_year = time.cell(0).point.timetuple()[0]
    np_indices = np.zeros((max_doy, no_years, window_size), int)
    for c in time.cells():
        tt = c.point.timetuple()
        year = tt[0]
        day_of_year = tt[7] - 1
        if day_of_year >= max_doy:
            continue
        idx_y = year - first_year
        days = np.arange(day_of_year - window_width, day_of_year + window_width + 1)
        np_indices[day_of_year, idx_y] = window_width + idx_y * 365 + days
    np_indices[0, 0, :2] = window_width
    np_indices[1, 0, :1] = window_width
    np_indices[-1, -1, -2:] = np_indices[-1, -1, -3]
    np_indices[-2, -1, -1:] = np_indices[-2, -1, -2]
    return np_indices


class CountPercentileOccurrences(IndexFunction):
    def __init__(self, percentile, condition, reference_period, bootstrapping=True):
        super().__init__(units="%")
        self.reference_period = reference_period
        percentile.convert_units("1")
        self.percentile = float(percentile.points)
        self.condition = NUMPY_OPERATORS[condition]
        self.lazy_condition = DASK_OPERATORS[condition]
        self.bootstrapping = bootstrapping

    def preprocess(self, cubes, client):
        window_size = 5
        window_width = window_size // 2
        cube = cubes[0]
        reference_period = ReferencePeriod(cube, self.reference_period)
        self.extra_coords.append(reference_period.aux_coord.copy())
        max_doy = 365
        self.years = {
            y: np.arange(i * window_size, (i + 1) * window_size)
            for i, y in enumerate(
                range(
                    reference_period.first_year,
                    reference_period.last_year + 1,
                )
            )
        }
        logging.info("Building indices")
        time = cube.coord("time")
        try:
            np_indices = build_indices(
                time[reference_period.idx_0 : reference_period.idx_n + 1],
                max_doy,
                len(self.years),
                window_size,
            )
        except IndexError as e:
            raise IndexError(
                "Can only build indices for a reference period that includes "
                "full years, e.g. 1990-01-01/1999-12-31. Unable to build indices "
                "for a subset of years, e.g. 1990-05-01/1999-04-30"
            ) from e
        logging.info("Arranging data")
        start_index = reference_period.idx_0 - window_width
        end_index = reference_period.idx_n + window_width + 1
        all_data = da.moveaxis(
            cube.core_data()[start_index:end_index],
            0,
            -1,
        )
        all_data = all_data.rechunk(("auto",) * (all_data.ndim - 1) + (-1,))
        all_data = client.persist(all_data)
        progress(all_data)
        data = []
        for idx_d in range(max_doy):
            d = client.persist(all_data[..., np_indices[idx_d].ravel()])
            progress(d)
            data.append(d)
            logging.info(f"Finished doy {idx_d}")
        data = da.stack(data, axis=0)
        data = data.rechunk({0: "auto"})
        logging.info(f"data chunks: {data.chunks}")
        data = client.persist(data)
        progress(data)
        logging.info(f"data chunks: {data.chunks}")
        chunks = (-1,) + (10,) * (data.ndim - 2) + (-1,)
        data = data.rechunk(chunks)
        data = client.persist(data)
        progress(data)
        logging.info("Initializing quantiles")
        self.quantiler = BootstrapQuantiles(
            data,
            self.percentile,
            reference_period.first_year,
            window_size,
            client,
            axis=-1,
        )
        logging.info("Starting quantile calculation")
        res = client.persist(self.quantiler.quantiles().rechunk())
        progress(res)
        self.out_of_base_quantiles = res

    def call_func(self, data, axis, **kwargs):
        pass

    def lazy_func(self, data, axis, cube, client, **kwargs):
        year = cube.coord("time").cell(0).point.year
        logging.info(f"Starting year {year}")
        if data.shape[0] > 365:
            data = data[:-1]
        if self.bootstrapping and year in self.years:
            logging.info("Using bootstrapping")
            quantile_years = [y for y in self.years.keys() if y != year]
            counts = []
            for duplicate_year in quantile_years:
                quantiles = self.quantiler.quantiles(
                    ignore_year=year, duplicate_year=duplicate_year
                )
                cond = self.lazy_condition(data[...], quantiles)
                count = da.count_nonzero(cond, axis=0)
                counts.append(count)
            counts = da.stack(counts, axis=-1)
            counts = counts.rechunk(("auto",) * (counts.ndim - 1) + (-1,))
            avg_counts = counts.mean(axis=-1)
            percents = avg_counts / (data.shape[0] / 100.0)
        else:
            logging.info("Not using bootstrapping")
            cond = self.lazy_condition(data, self.out_of_base_quantiles)
            counts = da.count_nonzero(cond, axis=0).astype(np.float32)
            percents = counts / (data.shape[0] / 100.0)
        return percents


class CountThresholdedPercentileOccurrences(CountPercentileOccurrences, IndexFunction):
    def __init__(
        self,
        percentile,
        data_threshold,
        data_condition,
        percentile_condition,
        reference_period,
        **kwargs,
    ):
        super().__init__(percentile, percentile_condition, reference_period)
        self.data_threshold = data_threshold
        self.data_condition = NUMPY_OPERATORS[data_condition]
        self.data_lazy_condition = DASK_OPERATORS[data_condition]

    def preprocess(self, cubes, client):
        cube = cubes[0]
        logging.info("Convert threshold unit to match data unit")
        data_threshold = self.data_threshold
        data_threshold.points = data_threshold.points.astype(cube.dtype)
        if data_threshold.has_bounds():
            data_threshold.bounds = data_threshold.bounds.astype(cube.dtype)
        change_units(data_threshold, cube.units, cube.standard_name)
        reference_period = ReferencePeriod(cube, self.reference_period)
        self.extra_coords.append(reference_period.aux_coord.copy())
        logging.info("Arranging data")
        start_index = reference_period.idx_0
        end_index = reference_period.idx_n + 1
        ref_data = da.moveaxis(
            cube.core_data()[start_index:end_index],
            0,
            -1,
        )
        ref_data = ref_data.rechunk(("auto",) * (ref_data.ndim - 1) + (-1,))
        progress(ref_data)
        logging.info(f"data chunks: {ref_data.chunks}")
        logging.info("Initializing quantiles")
        quantiler = BootstrapQuantilesThreshold(
            ref_data,
            self.data_threshold.points,
            self.data_lazy_condition,
            self.percentile,
            365,
            client,
            axis=-1,
        )
        logging.info("Starting quantile calculation")
        res = client.persist(quantiler.quantiles().rechunk())
        progress(res)
        res = client.persist(da.ma.filled(res, fill_value=0.0))
        progress(res)
        self.out_of_base_quantiles = res

    def call_func(self, data, axis, **kwargs):
        self.axis = normalize_axis(axis, data.ndim)
        data_mask = np.ma.getmaskarray(data)
        data_mask = data_mask.any(axis=axis)
        cond = self.data_condition(data, self.data_threshold.points)
        threshold_cond = np.ma.masked_where(~cond, data)
        percentile_cond = self.condition(threshold_cond, self.out_of_base_quantiles)
        masked_data = np.ma.masked_where(~percentile_cond, data)
        output_unit = kwargs["output_metadata"].units
        res = convert_to_desired_output(
            self, threshold_cond, masked_data, axis, output_unit
        )
        res = np.ma.filled(res, fill_value=0.0)
        res = np.ma.masked_where(data_mask, res)
        return res.astype(np.float32)

    def lazy_func(self, data, axis, **kwargs):
        self.axis = normalize_axis(axis, data.ndim)
        data_mask = da.ma.getmaskarray(data)
        data_mask = data_mask.any(axis=axis)
        cond = self.data_lazy_condition(data, self.data_threshold.points)
        threshold_cond = da.ma.masked_where(~cond, data)
        percentile_cond = self.lazy_condition(
            threshold_cond, self.out_of_base_quantiles
        )
        masked_data = da.ma.masked_where(~percentile_cond, data)
        output_unit = kwargs["output_metadata"].units
        res = convert_to_desired_output(
            self, threshold_cond, masked_data, axis, output_unit
        )
        res = da.ma.filled(res, fill_value=0.0)
        res = da.ma.masked_where(data_mask, res)
        return res.astype(np.float32)



#!/usr/bin/python3

import re
import sys


def wait_for_enter():
    input("Press Enter to continue: ")
    print("")


class Prepare:
    def run(self, context):
        print(
            (
                "> Place yourself in the root folder of your climix fork clone,"
                " checkout the 'main' branch and do 'git fetch --all'."
            )
        )
        wait_for_enter()
        print(
            (
                "> Make sure your 'main' branch is up to date with the upstream"
                " 'main' and that there are no local changes."
            )
        )
        wait_for_enter()
        print(
            (
                "> Make sure the dependency list in 'pyproject.toml' is up to date"
                " with the non-developer dependencies in the 'environment.yml'"
                " file. Also make sure the other sections of 'project.toml' file"
                " is up to date."
            )
        )
        wait_for_enter()


class UpdateChangeLog:
    def run(self, context):
        print(
            f"> Run the command 'utils/update-changelog.sh"
            f" {context['new_version'][1:]}'."
        )
        wait_for_enter()
        print(
            (
                "> Copy the output from the command and paste it into"
                " 'CHANGELOG.md' (the structure of the document will show where"
                " to put the generated text)."
            )
        )
        wait_for_enter()
        print(
            (
                "> In the pasted text in 'CHANGELOG.md', move each commit to a"
                " proper commit category section"
            )
        )
        wait_for_enter()
        print(
            (
                "> In the pasted text in 'CHANGELOG.md', remove commit category"
                " headers with no commits."
            )
        )
        wait_for_enter()
        print(
            (
                "> At the end of 'CHANGELOG.md', replace the top line of the link"
                " list (the line starting with [unreleased]) with the following"
                " two lines:"
            )
        )
        print(
            f"[unreleased]: https://git.smhi.se/climix/climix/compare/"
            f"{context['new_version']}...HEAD"
        )
        print(
            f"[{context['new_version'][1:]}]:"
            f" https://git.smhi.se/climix/climix/compare/"
            f"{context['old_version']}...{context['new_version']}"
        )
        wait_for_enter()


class CreateNewBranchAndCommit:
    def run(self, context):
        print(
            f"> Create a new branch called 'changelog-{context['new_version']}'."
            f" and checkout the new branch."
        )
        wait_for_enter()
        print(
            f"> Add the 'CHANGELOG.md' file and make a new commit with message:"
            f"'Update changelog for release {context['new_version']}'."
        )
        wait_for_enter()
        print("> Push the new branch to your climix fork.")
        wait_for_enter()


class CreateMergeRequestAndMerge:
    def run(self, context):
        print(
            (
                "> Go to https://git.smhi.se/climix/climix and create a new"
                " merge request for the new branch from your fork into the"
                " main branch of the climix upstream repo."
            )
        )
        wait_for_enter()
        print(("> Have a colleague sanity check the merge request."))
        wait_for_enter()
        print(("> Merge."))
        wait_for_enter()


class TagVersion:
    def run(self, context):
        print(
            (
                "> Checkout the 'main' branch of your fork, and update it to"
                " be in sync with the upstream repo."
            )
        )
        wait_for_enter()
        print(
            f"> Run the command 'git tag -a {context['new_version']}',"
            f" and use the tag message 'Release {context['new_version']}'."
        )
        wait_for_enter()
        print(
            f"> Push the new tag by running the command"
            f" 'git push upstream {context['new_version']}'."
        )
        wait_for_enter()


class BuildForPYPI:
    def run(self, context):
        print(
            (
                "> This section requires an account on https://pypi.org/, and"
                " access to the pypi climix project. It also requires an account"
                " on https://readthedocs.org and access to the readthedocs climix"
                " project. If you don't have this, please talk to a colleague."
            )
        )
        wait_for_enter()
        print(
            (
                "> If you have a 'dist' folder in the root of your clone, run"
                " the command 'rm dist/*' to clear out old builds."
            )
        )
        wait_for_enter()
        print(
            (
                "> Make sure to activate a mamba environment with updated"
                " versions of 'pip' and 'twine' installed."
            )
        )
        wait_for_enter()
        print(
            (
                "> Run the command 'python -m build', and check that you have"
                " a new build in the 'dist' folder."
            )
        )
        wait_for_enter()
        print(
            (
                "> Make sure you have an api token from pypi. This is needed to"
                " upload the new build."
            )
        )
        wait_for_enter()
        print(("> From your clone root folder, run the command 'twine upload dist/*'."))
        wait_for_enter()
        print(("> Make sure the new build is available at https://pypi.org/."))
        wait_for_enter()
        print(
            (
                "> Make sure the documentation for the new version is available"
                " at https://climix.readthedocs.io/ (may take a moment after pypi"
                " upload for the docs to get updated)."
            )
        )
        wait_for_enter()


class UpdateOnConda:
    def run(self, context):
        print(
            (
                "> This section requires an account on github, and access to"
                " the climix conda-forge feedstock repo. If you don't have this,"
                " please talk to a colleague."
            )
        )
        wait_for_enter()
        print(("> Go to https://github.com/conda-forge/climix-feedstock."))
        wait_for_enter()
        print(
            (
                "> In the file 'recipe/meta.yaml', make sure the 'requirements'"
                " list is up to date with the 'dependencies' list in the climix"
                " repo 'pyproject.toml' file. You can edit and commit changes"
                " to 'recipe/meta.yaml' in the browser if necessary."
            )
        )
        wait_for_enter()
        print(
            (
                "> Create a new issue, then choose the 'Bot commands' category."
                " Set the issue title to"
                " '@conda-forge-admin, please update version'."
            )
        )
        wait_for_enter()
        print(
            (
                "> conda-forge-webservices will create a pull request and"
                " run tests. Wait for it to finish."
            )
        )
        wait_for_enter()
        print(
            (
                "> After all tests have finished successfully, merge the"
                " merge request."
            )
        )
        wait_for_enter()
        print(
            (
                "> Make sure the new version appears on "
                " https://anaconda.org/conda-forge/climix. It may take a short "
                " while after the merge."
            )
        )
        wait_for_enter()


def _checkargs():
    if len(sys.argv) < 3:
        print(
            (
                "Please provide the previous and current release version"
                " numbers, in format 'vMajor.Minor.Patch',"
            )
        )
        print(("e.g.: 'python3 climix-release.py v0.20.0 v0.21.0'"))
        return False

    version_re = r"^v[0-9]+.[0-9]+.[0-9]+"
    old_version_ok = re.fullmatch(version_re, sys.argv[1])
    new_version_ok = re.fullmatch(version_re, sys.argv[2])

    if not old_version_ok or not new_version_ok:
        print(
            (
                "Format number in wrong format. Format should be"
                " 'vMajor.Minor.Patch', e.g. 'v0.20.0'."
            )
        )
        return False

    return True


if __name__ == "__main__":
    if _checkargs():
        context = {
            "old_version": sys.argv[1],
            "new_version": sys.argv[2],
        }

        procedure = [
            Prepare(),
            UpdateChangeLog(),
            CreateNewBranchAndCommit(),
            CreateMergeRequestAndMerge(),
            TagVersion(),
            BuildForPYPI(),
            UpdateOnConda(),
        ]
        for step in procedure:
            print("----------------------------------------")
            print(step.__class__.__name__)
            print("----------------------------------------")
            step.run(context)
        print("Done.")



# -*- coding: utf-8 -*-
import logging
from collections import namedtuple
from datetime import (
    datetime,
    time,
)
import isodate
import operator
import re

from cf_units import Unit
import cftime
import dask.array as da
import iris
import numpy as np

from ..util import change_units


SUPPORTED_OPERATORS = [
    "<",
    ">",
    "<=",
    ">=",
]
SUPPORTED_REDUCERS = [
    "min",
    "max",
    "sum",
    "mean",
]


def scalar_mean(*args):
    return float(np.mean(args))


SCALAR_OPERATORS = {
    "<": operator.lt,
    ">": operator.gt,
    "<=": operator.le,
    ">=": operator.ge,
}
SCALAR_REDUCERS = {
    "min": min,
    "max": max,
    "sum": sum,
    "mean": scalar_mean,
}


NUMPY_OPERATORS = {
    "<": np.less,
    ">": np.greater,
    "<=": np.less_equal,
    ">=": np.greater_equal,
}
NUMPY_REDUCERS = {
    "min": np.min,
    "max": np.max,
    "sum": np.sum,
    "mean": np.mean,
}


DASK_OPERATORS = {
    "<": da.less,
    ">": da.greater,
    "<=": da.less_equal,
    ">=": da.greater_equal,
}
DASK_REDUCERS = {
    "min": da.min,
    "max": da.max,
    "sum": da.sum,
    "mean": da.mean,
}


TimeRange = namedtuple("TimeRange", ["start", "end", "climatological"])


TIME_FORMATS = {
    4: "%Y",
    6: "%Y%m",
    8: "%Y%m%d",
    10: "%Y%m%d%H",
    12: "%Y%m%d%H%M",
    14: "%Y%m%d%H%M%S",
}


def parse_timerange(timerange):
    parts = timerange.split("-")
    n = len(parts)
    if n < 2 or n > 3:
        raise ValueError(f"Invalid timerange {timerange}")
    if n == 3:
        if parts[2] != "clim":
            raise ValueError(f"Invalid timerange {timerange}")
        climatological = True
    else:
        climatological = False
    n_start = len(parts[0])
    n_end = len(parts[1])
    if n_start != n_end:
        raise ValueError(
            f"Start and end time must have the same resolution in {timerange}"
        )
    format_string = TIME_FORMATS[n_start]
    start_time = datetime.strptime(parts[0], format_string)
    end_time = datetime.strptime(parts[1], format_string)
    return TimeRange(start_time, end_time, climatological)


def parse_isodate(date):
    """
    Parses a date string in ISO 8601 format and
    return a datetime object.
    """
    parsed_date = isodate.parse_date(date)
    return datetime.combine(parsed_date, time.min)


def parse_isodate_upper_bound(date):
    """
    Parses a four-digit year or calendar date string in ISO8601 format
    and returns the upper bound for the time interval defined by the
    precision of the given date (i.e. the strings `1990`, `1990-12` and
    `1990-12-31` will all result in a datetime object with the time
    `1991-01-01T00:00:00`).
    """
    parsed_date = isodate.parse_date(date)
    if re.fullmatch(r"[0-9]{4}", date):  # YYYY
        parsed_date = parsed_date + isodate.Duration(years=1)
    elif re.fullmatch(r"[0-9]{4}-[0-9]{2}", date):  # YYYY-MM
        parsed_date = parsed_date + isodate.Duration(months=1)
    elif re.fullmatch(r"[0-9]{4}-[0-9]{2}-[0-9]{2}", date):  # YYYY-MM-DD
        parsed_date = parsed_date + isodate.Duration(days=1)
    else:
        raise ValueError(
            "Only four-digit years and calendar dates in ISO8601 format are supported."
        )
    return datetime.combine(parsed_date, time.min)


def parse_isodate_timerange(timerange):
    """
    Parses a time interval in ISO8601 format.
    """
    try:
        start, end = timerange.split("/")
    except ValueError as e:
        raise ValueError(
            "timerange string must contain one and only one slash ('/')"
        ) from e
    if start.startswith("P"):
        end_date = parse_isodate_upper_bound(end)
        start_date = end_date - isodate.parse_duration(start)
    elif end.startswith("P"):
        start_date = parse_isodate(start)
        end_date = start_date + isodate.parse_duration(end)
    else:
        start_date = parse_isodate(start)
        end_date = parse_isodate_upper_bound(end)
    if start_date >= end_date:
        raise ValueError("Start date must be earlier than end date.")
    return TimeRange(start_date, end_date, False)


class TimesHelper:
    def __init__(self, time):
        self.times = time.core_points()
        self.units = str(time.units)

    def __getattr__(self, name):
        return getattr(self.times, name)

    def __len__(self):
        return len(self.times)

    def __getitem__(self, key):
        return self.times[key]


class ComputationalPeriod:
    """
    Builds a :namedtuple:`TimeRange` for the computational period.

    Given a time period following the ISO8601 format, returns the
    :namedtuple:`TimeRange` containing datetime objects for the start and end.
    """

    def __init__(self, computational_period):
        if computational_period is not None:
            self.timerange = parse_isodate_timerange(computational_period)
        else:
            self.timerange = None


class ReferencePeriod:
    """
    Builds the reference period from the time coordinate and the timerange. Stores
    information about first and last year, start and end index for the cube data,
    and a auxiliary coordinate with information.
    """

    def __init__(self, cube, reference_period):
        self.timerange = parse_isodate_timerange(reference_period)
        self.first_year = self.timerange.start.year
        self.last_year = self.timerange.end.year - 1
        logging.info(f"Building reference period <{reference_period}>")
        time = cube.coord("time")
        indices, bounds = self._get_indices_and_bnds(time, self.timerange)
        self.idx_0, self.idx_n = indices
        bnds_start, bnds_end = bounds
        middle_point = (bnds_start + bnds_end) / 2
        bounds = np.array([bnds_start, bnds_end])
        self.aux_coord = iris.coords.AuxCoord(
            var_name="reference_period",
            standard_name="reference_epoch",
            points=middle_point,
            bounds=bounds,
            units=time.units,
        )

    def _compare_data_period_bnds(self, time, bnds_start, bnds_end):
        """
        Check that the boundaries does not exceed the time range.
        """
        if bnds_start < time.bounds[0][0] or (bnds_end - 1) > time.bounds[-1][1]:
            start_date = cftime.num2date(
                time.bounds[0][0], time.units.name, calendar=time.units.calendar
            )
            end_date = cftime.num2date(
                time.bounds[-1][1], time.units.name, calendar=time.units.calendar
            )
            raise ValueError(
                "The reference period cannot be outside of the data time range. "
                f"Choose a reference period between <{start_date}, {end_date}>"
            )

    def _get_indices_and_bnds(self, time, timerange):
        """
        Returns the start and end index and the boundary points in relation to the cube
        time coordinate
        """
        time_units = time.units
        times = TimesHelper(time)
        bnds_start = cftime.date2num(
            timerange.start, time_units.name, calendar=time_units.calendar
        )
        bnds_end = cftime.date2num(
            timerange.end, time_units.name, calendar=time_units.calendar
        )
        self._compare_data_period_bnds(time, bnds_start, bnds_end)
        idx_0 = cftime.date2index(
            timerange.start, times, calendar=time_units.calendar, select="after"
        )
        idx_n = cftime.date2index(
            timerange.end, times, calendar=time_units.calendar, select="before"
        )
        if time[idx_n].points[0] == bnds_end:
            # If the end date exists exactly in the data, don't include it.
            # E.g., reference period "2014/2018" should have bounds
            # [2014-01-01T00:00:00, 2019-01-01T00:00:00] and the indexing of the data
            # should be the first time step after 2014-01-01T00:00:00 and the last time
            # step before 2019-01-01T00:00:00.
            idx_n -= 1
        return (idx_0, idx_n), (bnds_start, bnds_end)


def normalize_axis(axis, ndim):
    if isinstance(axis, list) and len(axis) == 1:
        axis = axis[0]
    if axis < 0:
        # just cope with negative axis numbers
        axis += ndim
    return axis


class IndexFunction:
    def __init__(self, standard_name=None, units=Unit("no_unit")):
        super().__init__()
        self.standard_name = standard_name
        self.units = units
        self.extra_coords = []

    def preprocess(self, cubes, client):
        pass

    def prepare(self, input_cubes):
        pass

    def pre_aggregate_shape(self, *args, **kwargs):
        return ()


class ThresholdMixin:
    def __init__(self, threshold, condition, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.threshold = threshold
        self.condition = NUMPY_OPERATORS[condition]
        self.lazy_condition = DASK_OPERATORS[condition]
        self.extra_coords.append(threshold.copy())

    def prepare(self, input_cubes):
        ref_cube = next(iter(input_cubes.values()))
        threshold = self.threshold
        threshold.points = threshold.points.astype(ref_cube.dtype)
        if threshold.has_bounds():
            threshold.bounds = threshold.bounds.astype(ref_cube.dtype)
        change_units(threshold, ref_cube.units, ref_cube.standard_name)
        super().prepare(input_cubes)


class ReducerMixin:
    def __init__(self, reducer, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reducer = NUMPY_REDUCERS[reducer]
        self.lazy_reducer = DASK_REDUCERS[reducer]
        self.scalar_reducer = SCALAR_REDUCERS[reducer]


class RollingWindowMixin:
    def __init__(self, rolling_aggregator, window_size, reducer, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reducer = NUMPY_REDUCERS[reducer]
        self.lazy_reducer = DASK_REDUCERS[reducer]
        self.window_size = window_size
        self.rolling_aggregator = NUMPY_REDUCERS[rolling_aggregator]
        self.lazy_rolling_aggregator = DASK_REDUCERS[rolling_aggregator]



import pytest

from climix.index_functions import support
from datetime import datetime


TEST_PARSE_ISODATE_PARAMETERS = [
    ("1961", datetime(1961, 1, 1)),
    ("2050", datetime(2050, 1, 1)),
    ("2010-01", datetime(2010, 1, 1)),
    ("2010-07", datetime(2010, 7, 1)),
    ("2010-01-01", datetime(2010, 1, 1)),
    ("2010-07-26", datetime(2010, 7, 26)),
]


@pytest.mark.parametrize("date, expected", TEST_PARSE_ISODATE_PARAMETERS)
def test_parse_isodate(date, expected):
    """Test ``test_parse_isodate``."""
    assert support.parse_isodate(date) == expected


TEST_PARSE_ISODATE_UPPER_BOUND_PARAMETERS = [
    ("1961", datetime(1962, 1, 1)),
    ("2050", datetime(2051, 1, 1)),
    ("2010-01", datetime(2010, 2, 1)),
    ("2010-07", datetime(2010, 8, 1)),
    ("2010-01-01", datetime(2010, 1, 2)),
    ("2010-07-26", datetime(2010, 7, 27)),
]


@pytest.mark.parametrize("date, expected", TEST_PARSE_ISODATE_UPPER_BOUND_PARAMETERS)
def test_parse_isodate_upper_bound(date, expected):
    """Test ``test_parse_isodate_upper_bound``."""
    assert support.parse_isodate_upper_bound(date) == expected


TEST_PARSE_ISODATE_TIMERANGE_PARAMETERS = [
    ("1961/1990", support.TimeRange(datetime(1961, 1, 1), datetime(1991, 1, 1), False)),
    ("2010/2050", support.TimeRange(datetime(2010, 1, 1), datetime(2051, 1, 1), False)),
    (
        "1961-01/2050-01",
        support.TimeRange(datetime(1961, 1, 1), datetime(2050, 2, 1), False),
    ),
    (
        "1961-01/2050-01",
        support.TimeRange(datetime(1961, 1, 1), datetime(2050, 2, 1), False),
    ),
    (
        "1961-01/2050-01",
        support.TimeRange(datetime(1961, 1, 1), datetime(2050, 2, 1), False),
    ),
    ("P20Y/1990", support.TimeRange(datetime(1971, 1, 1), datetime(1991, 1, 1), False)),
    ("1961/P20Y", support.TimeRange(datetime(1961, 1, 1), datetime(1981, 1, 1), False)),
    (
        "P20Y6M/1990",
        support.TimeRange(datetime(1970, 7, 1), datetime(1991, 1, 1), False),
    ),
    (
        "1961/P20Y6M",
        support.TimeRange(datetime(1961, 1, 1), datetime(1981, 7, 1), False),
    ),
]


@pytest.mark.parametrize("timerange, expected", TEST_PARSE_ISODATE_TIMERANGE_PARAMETERS)
def test_parse_isodate_timerange(timerange, expected):
    """Test ``test_parse_isodate_timerange``."""
    assert support.parse_isodate_timerange(timerange) == expected



from collections import namedtuple

import numpy as np
from numba import jit

Kernels = namedtuple(
    "Kernels", ["chunk", "aggregate", "combine", "post_process"], defaults=[None, None]
)


def make_first_spell_kernels():
    # The gufunc support in numba is lacking right now.
    # Once numba supports NEP-20 style signatures for
    # gufunc, we can use @guvectorize to vectorize this
    # function, allowing for more general applicability
    # in terms of dimensions and hopefully better performance,
    # perhaps taking advantage of GPUs. The template is
    # @guvectorize([float32, int64], '(n)->(4)')
    # def chunk_column(column, res):
    @jit(nopython=True)
    def chunk_column(column, minimum_length):
        """Calculate first spell information for a single timeseries.

        Parameters
        ----------
        column : np.array
            1d array containing the timeseries

        Returns
        -------
        np.array
            4 vector containing the first spell information, namely:
            - length of timeseries
            - length of spell at beginning
            - position of first spell in the timeseries
            - length of spell at the end
        """
        res = np.empty((4,), dtype=np.float32)
        n = column.shape[0]
        where = np.flatnonzero
        x = column[1:] != column[:-1]
        starts = where(x) + 1
        no_runs = len(starts)
        # the length of the chunk is n
        res[0] = n
        # assume no internal spell
        res[2] = np.nan
        # if no change occurs, then...
        if no_runs == 0:
            # ...either the spell covers the entire chunk, or...
            if column[0]:
                res[1] = n
                if n >= minimum_length:
                    res[2] = 0
                res[3] = n
            # ...there is no spell in this chunk
            else:
                res[1] = 0
                res[3] = 0
        else:
            # there is a spell from the beginning to the first change
            # if the first value is part of a spell
            res[1] = starts[0] if column[0] else 0
            # there is a spell from the last change to the end if
            # the last value is part of a spell
            res[3] = n - starts[-1] if column[starts[-1]] else 0
            # if there is a sufficiently long spell at the beginning, we're done
            if res[1] >= minimum_length:
                res[2] = 0
                return res
            # if there is a sufficiently long spell at the end,
            # we don't need to look at more chunks in the aggregation
            if res[3] >= minimum_length:
                res[2] = starts[-1]
            # if there are at least two changes (range(1) = [0])
            for k in range(no_runs - 1):
                # if the value at the corresponding change is part
                # of a spell then this spell stretches to the next
                # change and is an internal spell
                if column[starts[k]]:
                    length = starts[k + 1] - starts[k]
                    if length >= minimum_length:
                        res[2] = starts[k]
                        return res
        return res

    @jit(nopython=True)
    def find_edge_spell(thresholded_data, cond, i, minimum_length):
        # find where the spell is valid
        offset_data = np.logical_and(cond, thresholded_data)
        # find where the spell ends
        spell_lengths = np.argwhere(offset_data == 0)
        # if the spell ends within the chunk
        if len(spell_lengths) > 0:
            # if there is a start spell
            if i == 0:
                res = spell_lengths[0][0]
            # if there is a end spell
            else:
                res = offset_data.shape[-1] - spell_lengths[-1][0] - 1
        else:
            # if the spell covers the duration
            if thresholded_data.shape[-1] > minimum_length:
                res = minimum_length - 1
            # if the chunk is smaller than the duaration
            # and the spell covers the chunk
            else:
                res = thresholded_data.shape[-1]
        return res

    @jit(nopython=True)
    def chunk(thresholded_data, offset, chunk_time, length, minimum_length):
        res = np.empty(thresholded_data.shape[:-1] + (6,), dtype=np.float32)
        # if the chunk contains the start/end of the year (edge spell)
        start_spell = np.where(chunk_time < (minimum_length - 1), True, False)
        end_spell = np.where(chunk_time > (length - minimum_length), True, False)
        # for each gridcell find the start spell and edge spells
        for ind in np.ndindex(*thresholded_data.shape[:-1]):
            offset_data = np.logical_and(offset, thresholded_data[ind])
            res[ind][:-2] = chunk_column(offset_data, minimum_length)
            for i, cond in enumerate([start_spell, end_spell]):
                # if there is a spell in the beginning/end of the year
                if cond.any():
                    res[ind][-(2 - i)] = find_edge_spell(
                        thresholded_data[ind], cond, i, minimum_length
                    )
                # otherwise the edge spell is 0
                else:
                    res[ind][-(2 - i)] = 0
        return res

    @jit(nopython=True)
    def combine(x_chunk, start_duration, end_duration):
        def _merge_chunks(chunks, this, minimum_length):
            this = this.copy()
            # mark where this chunk is completely covered by a spell
            this_full = np.asarray(this[..., 0] == this[..., 1])
            for k in range(1, chunks.shape[0]):
                next_chunk = chunks[k]
                for ind in np.ndindex(this.shape[:-1]):
                    ind_length = ind + (0,)
                    ind_head = ind + (1,)
                    ind_internal = ind + (2,)
                    ind_tail = ind + (3,)
                    ind_spell_start = ind + (4,)
                    ind_spell_end = ind + (5,)
                    # if the start edge spell covers the chunk
                    start_full = np.asarray(this[ind_spell_start] == this[ind_length])
                    if start_full:
                        this[ind_spell_start] += next_chunk[ind_spell_start]
                    # if the end edge spell covers the next chunk
                    end_full = np.asarray(
                        next_chunk[ind_spell_end] == next_chunk[ind_length]
                    )
                    if end_full:
                        this[ind_spell_end] += next_chunk[ind_spell_end]
                    else:
                        this[ind_spell_end] = next_chunk[ind_spell_end]
                    # the next chunk is completely covered by a spell
                    next_full = next_chunk[ind_length] == next_chunk[ind_head]
                    # if both are completely covered the merged chunk
                    # is completely covered too
                    if this_full[ind] and next_full:
                        this[ind_head] += next_chunk[ind_head]
                        this[ind_tail] += next_chunk[ind_tail]
                        if this[ind_head] >= minimum_length:
                            this[ind_internal] = 0
                    # if the old chunk is completely covered, but the new one
                    # isn't, then
                    elif this_full[ind]:
                        # the head is the old head + the new head,
                        this[ind_head] += next_chunk[ind_head]
                        # the internal spell is the new internal spell,
                        if this[ind_head] >= minimum_length:
                            this[ind_internal] = 0
                        else:
                            this[ind_internal] = (
                                this[ind_length] + next_chunk[ind_internal]
                            )
                        # the tail is the new tail,
                        this[ind_tail] = next_chunk[ind_tail]
                        # and the resulting chunk is no longer fully covered
                        this_full[ind] = False
                    # if the old chunk is not fully covered, but the new one is
                    elif next_full:
                        old_tail = this[ind_tail]
                        # the tail is the old tail + the new head
                        this[ind_tail] += next_chunk[ind_head]
                        if this[ind_tail] >= minimum_length and np.isnan(
                            this[ind_internal]
                        ):
                            this[ind_internal] = this[ind_length] - old_tail
                    # if neither are fully covered
                    else:
                        # the head stays the same,
                        # the internal spell is the winner between
                        # the old internal spell, the new internal spell,
                        # and the internal spell resulting from merging
                        # old tail and new head,
                        if np.isnan(this[ind_internal]):
                            length = this[ind_tail] + next_chunk[ind_head]
                            if length >= minimum_length:
                                this[ind_internal] = this[ind_length] - this[ind_tail]
                            elif not np.isnan(next_chunk[ind_internal]):
                                this[ind_internal] = (
                                    this[ind_length] + next_chunk[ind_internal]
                                )
                        # and the tail is the new tail
                        this[ind_tail] = next_chunk[ind_tail]
                    # the length of the combined chunks is the sum of the
                    # lengths of the individual chunks
                    this[ind_length] += next_chunk[ind_length]
            return this

        def _find_overlap(chunks, this, ind, minimum_length):
            this = this.copy()
            length = chunks[(0,) + ind + (0, 0)]
            ind_length = (0, 0)
            ind_internal_start = (2, 0)
            ind_internal_end = (2, 1)
            ind_internal_end_dep = (2, 2)
            ind_head_end = (1, 1)
            ind_tail_start = (3, 0)
            ind_tail_end = (3, 1)
            start_ind = this[ind_internal_start]
            # combine the chunk results by merging them pairwise
            # trying to find a new end after the start
            for k in range(1, chunks.shape[0]):
                current = chunks[k - 1][ind]
                next_chunk = chunks[k][ind]
                # if the start is in the current chunk
                if length >= start_ind and not this[ind_internal_end_dep] > start_ind:
                    # if there is a end between the chunks
                    if (
                        current[ind_tail_end] + next_chunk[ind_head_end]
                        >= minimum_length
                    ):
                        new_tail = current[ind_tail_start] - 1
                        # if the end is after the start
                        if length - current[ind_tail_end] > start_ind:
                            this[ind_internal_end_dep] = length - current[ind_tail_end]
                        # if there is a valid end after the start
                        elif new_tail + next_chunk[ind_head_end] >= minimum_length:
                            this[ind_internal_end_dep] = length - new_tail
                        else:
                            # if there is a independent end in the next chunk
                            if length + next_chunk[ind_internal_end] > start_ind:
                                this[ind_internal_end_dep] = (
                                    length + next_chunk[ind_internal_end]
                                )
                    # the end might be in the next chunk
                    else:
                        # there is an independent end
                        if length + next_chunk[ind_internal_end] > start_ind:
                            this[ind_internal_end_dep] = (
                                length + next_chunk[ind_internal_end]
                            )
                # if the start is in the next chunk
                elif (
                    length + next_chunk[ind_length] >= start_ind
                    and not this[ind_internal_end_dep] > start_ind
                ):
                    # if there is a independent end after the start
                    if length + next_chunk[ind_internal_end] > start_ind:
                        this[ind_internal_end_dep] = (
                            length + next_chunk[ind_internal_end]
                        )
                    # if there is a dependent end after the start
                    elif length + next_chunk[ind_internal_end_dep] > start_ind:
                        this[ind_internal_end_dep] = (
                            length + next_chunk[ind_internal_end_dep]
                        )
                length += next_chunk[ind_length]
            # if the start is between chunks
            # there might be an dependent end before the start
            if this[ind_internal_end_dep] <= start_ind:
                this[ind_internal_end_dep] = np.nan
            return this

        # if there is only one chunk
        res = x_chunk[0].copy()
        # if there are more than one chunk
        if x_chunk.shape[0] > 1:
            # find start
            res[..., 0] = _merge_chunks(x_chunk[..., 0], res[..., 0], start_duration)
            # if we have conditions for finding a end
            if not np.isnan(res[..., 0, 1]).any():
                # find end independent of start
                res[..., 1] = _merge_chunks(x_chunk[..., 1], res[..., 1], end_duration)
                # if the end index is before or at the same time as the start index
                # update the dependent end
                for ind in np.ndindex(res.shape[:-2]):
                    ind_start = ind + (2, 0)
                    ind_end = ind + (2, 1)
                    if res[ind_start] >= res[ind_end]:
                        res[ind] = _find_overlap(x_chunk, res[ind], ind, end_duration)
        return res

    @jit(nopython=True)
    def aggregate(res):
        for ind in np.ndindex(res.shape[:-2]):
            ind_start = ind + (2, 0)
            ind_end = ind + (2, 1)
            ind_end_dep = ind + (2, 2)
            # if there is no start there is no end
            if np.isnan(res[ind_start]):
                res[ind_end] = np.nan
            # if the independent end is before or at the same time as the start
            # set the end to the dependent end
            elif res[ind_start] >= res[ind_end]:
                res[ind_end] = res[ind_end_dep]
            res[ind_start] += 1
            res[ind_end] += 1
        return res

    return Kernels(chunk, aggregate, combine)


def make_spell_length_kernels(reducer):
    # The gufunc support in numba is lacking right now.
    # Once numba supports NEP-20 style signatures for
    # gufunc, we can use @guvectorize to vectorize this
    # function, allowing for more general applicability
    # in terms of dimensions and hopefully better performance,
    # perhaps taking advantage of GPUs. The template is
    # @guvectorize([float32, int64], '(n)->(4)')
    # def chunk_column(column, res):
    @jit(nopython=True)
    def chunk_column(column):
        res = np.empty((7,), dtype=np.float32)
        n = column.shape[0]
        where = np.flatnonzero
        x = column[1:] != column[:-1]
        starts = where(x) + 1
        no_runs = len(starts)
        # the length of the chunk is n
        res[0] = n
        # assume no internal spell
        res[2] = 0
        # no start for any spell
        res[4:] = np.nan
        # if no change occurs, then...
        if no_runs == 0:
            # ...either the spell covers the entire chunk, or...
            if column[0]:
                res[1] = n
                res[3] = n
                res[4] = 1
                res[6] = 1
            # ...there is no spell in this chunk
            else:
                res[1] = 0
                res[3] = 0
        else:
            # there is a spell from the beginning to the first change
            # if the first value is part of a spell
            res[1] = starts[0] if column[0] else 0
            res[4] = 1 if column[0] else np.nan
            # there is a spell from the last change to the end if
            # the last value is part of a spell
            res[3] = n - starts[-1] if column[starts[-1]] else 0
            res[6] = starts[-1] + 1 if column[starts[-1]] else np.nan
            # if there are at least two changes (range(1) = [0])
            for k in range(no_runs - 1):
                # if the value at the corresponding change is part
                # of a spell then this spell stretches to the next
                # change and is an internal spell
                if column[starts[k]]:
                    old_internal = res[2]
                    length = starts[k + 1] - starts[k]
                    res[2] = reducer(length, res[2])
                    res[5] = starts[k] + 1 if res[2] != old_internal else res[5]
        return res

    @jit(nopython=True)
    def chunk(thresholded_data):
        res = np.empty(thresholded_data.shape[:-1] + (7,), dtype=np.float32)
        for ind in np.ndindex(*thresholded_data.shape[:-1]):
            res[ind] = chunk_column(thresholded_data[ind])
        return res

    @jit(nopython=True)
    def aggregate(x_chunk):
        # start with the first chunk and merge all others subsequently
        res = x_chunk[0].copy()
        # mark where this chunk is completely covered by a spell
        this_full = np.asarray(res[..., 0] == res[..., 1])
        for k in range(1, x_chunk.shape[0]):
            next_chunk = x_chunk[k]
            for ind in np.ndindex(res.shape[:-1]):
                ind_length = ind + (0,)
                ind_head = ind + (1,)
                ind_internal = ind + (2,)
                ind_tail = ind + (3,)
                ind_start_internal = ind + (5,)
                ind_start_tail = ind + (6,)
                # the next chunk is completely covered by a spell
                next_full = next_chunk[ind_length] == next_chunk[ind_head]
                # the length of the combined chunks is the sum of the
                # lengths of the individual chunks
                old_length = res[ind_length]
                res[ind_length] += next_chunk[ind_length]
                # if both are completely covered the merged chunk
                # is completely covered too
                if this_full[ind] and next_full:
                    res[ind_head] += next_chunk[ind_head]
                    res[ind_tail] += next_chunk[ind_tail]
                # if the old chunk is completely covered, but the new one
                # isn't, then
                elif this_full[ind]:
                    # the head is the old head + the new head,
                    res[ind_head] += next_chunk[ind_head]
                    # the internal spell is the new internal spell,
                    res[ind_internal] = next_chunk[ind_internal]
                    res[ind_start_internal] = (
                        old_length + next_chunk[ind_start_internal]
                    )
                    # the tail is the new tail,
                    res[ind_tail] = next_chunk[ind_tail]
                    res[ind_start_tail] = old_length + next_chunk[ind_start_tail]
                    # and the resulting chunk is no longer fully covered
                    this_full[ind] = False
                # if the old chunk is not fully covered, but the new one is
                elif next_full:
                    # the tail is the old tail + the new head
                    res[ind_tail] += next_chunk[ind_head]
                    if np.isnan(res[ind_start_tail]):
                        res[ind_start_tail] = old_length + next_chunk[ind_start_tail]
                # if neither are fully covered
                else:
                    # the head stays the same,
                    # the internal spell is the winner between
                    # the old internal spell, the new internal spell,
                    # and the internal spell resulting from merging
                    # old tail and new head,
                    first = res[ind_internal]
                    second = res[ind_tail] + next_chunk[ind_head]
                    last = next_chunk[ind_internal]
                    res[ind_internal] = reducer(
                        first,
                        second,
                        last,
                    )
                    if first != res[ind_internal]:
                        if second == res[ind_internal]:
                            if np.isnan(res[ind_start_tail]):
                                res[ind_start_internal] = old_length + 1
                            else:
                                res[ind_start_internal] = res[ind_start_tail]
                        else:
                            res[ind_start_internal] = (
                                old_length + next_chunk[ind_start_internal]
                            )
                    # and the tail is the new tail
                    res[ind_tail] = next_chunk[ind_tail]
                    res[ind_start_tail] = old_length + next_chunk[ind_start_tail]
        return res

    return Kernels(chunk, aggregate)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
from argparse import RawTextHelpFormatter
import logging
import os
import re
import textwrap
import time

from gordias.dask_setup import setup_scheduler
from gordias.datahandling import prepare_input_data, save

from climix import __version__
from climix.metadata import load_metadata

LOG_LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}


def parse_args():
    parser = argparse.ArgumentParser(
        description=(f"A climate index package, version {__version__}."),
        formatter_class=RawTextHelpFormatter,
    )
    parser.add_argument(
        "-l",
        "--log-level",
        choices=LOG_LEVELS.keys(),
        default="info",
        help="the lowest priority level of log messages to display",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="write more detailed log messages"
    )
    parser.add_argument("-d", "--dask-scheduler", default="distributed-local-cluster")
    parser.add_argument(
        "-k",
        "--keep-open",
        action="store_true",
        help="keep climix running until key press (useful for debugging)",
    )
    parser.add_argument(
        "-p",
        "--period",
        help=textwrap.dedent(
            """\
            Specify period for index (overrides index default period).
            Must be one of annual,seasonal,monthly. For annual period,
            an optional argument can be given to specify a range of months,
            e.g. annual[jja] for multiple months, or annual[feb] for a
            single month.
            """
        ),
    )
    parser.add_argument(
        "-s",
        "--sliced-mode",
        action="store_true",
        help="activate calculation per period to avoid memory problems",
    )
    parser.add_argument(
        "-i",
        "--iterative-storage",
        action="store_true",
        help="store results iteratively per period",
    )
    parser.add_argument(
        "-o", "--output", dest="output_template", help="output filename"
    )
    parser.add_argument(
        "-f",
        "--metadata-file",
        metavar="METADATA_FILE",
        dest="metadata_files",
        action="append",
        help=textwrap.dedent(
            """\
            add an external metadata file (overrides any default definitions),
            can be used multiple times to add several files (the last specified
            file will override any earlier definitions)
            """
        ),
    ),
    parser.add_argument(
        "--activate-config",
        action="store_true",
        dest="activate_config",
        help=textwrap.dedent(
            """\
            specify if the climix configuration should be activated. This will apply
            the default configuration for the output global attributes metadata.
            The default configuration file can be overridden by adding an external
            climix config metadata file.
            """
        ),
    ),
    parser.add_argument(
        "-r",
        "--reference-period",
        help=textwrap.dedent(
            """\
                specify reference period for an index (overrides the default
                reference period in the index definition), accepted formats are
                ISO 8601.
                Examples of usage:
                -r 1971/2000
                -r 1971-06-01/2000
                -r P48Y/2000
                -r P47Y7M/2000
                --reference-period 1971/2000
            """
        ),
    )
    parser.add_argument(
        "-cp",
        "--computational-period",
        help=textwrap.dedent(
            """\
                Specify computational period for the index, i.e. limit the time
                interval to a subset of what is available in the input data.
                The start and end time strings must follow `yyyy-mm-dd/yyyy-mm-dd`
                format, e.g. 1961-01-01/1990-12-31.
            """
        ),
    )
    parser.add_argument(
        "-x",
        "--index",
        action="append",
        required=True,
        metavar="INDEX",
        dest="indices",
        help=textwrap.dedent(
            """\
            the index to calculate
            (use "-x list" to get a list of all available indices)
            """
        ),
    )
    parser.add_argument(
        "datafiles", nargs="*", metavar="DATAFILE", help="the input data files"
    )
    return parser.parse_args()


def setup_logging(log_level, verbose=False):
    if verbose:
        format = (
            "%(relativeCreated)8dms:%(filename)s:%(funcName)s() "
            "%(levelname)s:%(name)s:%(message)s"
        )
    else:
        format = "%(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(level=LOG_LEVELS[log_level], format=format)


def setup_configuration(index, configuration):
    if configuration:
        index_distribution = index.metadata.distribution
        extra_attributes = {
            "CLIMIX_VERSION": f"climix-{__version__}",
            "INDEX_DISTRIBUTION": (
                f"{index_distribution.name}-{index_distribution.version}"
                if index_distribution is not None
                and index_distribution.name is not None
                else ""
            ),
        }
        configuration["global_attributes"].extra_attributes = extra_attributes


def guess_output_template(datafiles, computational_period=None):
    """
    Form an output filename template based on the names of the input files.

    The approach is as follows. The basenames of the input paths are stripped
    of the file extension and then split at underscores ('_'). The first part,
    usually the variable name, is discarded. The last part, usually a
    timerange, is reformatted to ensure hyphenated form (`start-end`) even in
    the case of a single timestamp, for example a single year. All the other,
    middle parts are reassambled with underscores and are called `base`.

    If the input files are both historical and scenario simulations the respective
    bases are expected to be identical except for one occurrence of a keyword from
    a list of known historical and scenario keywords. The occurring keywords
    will be concatenated with a hyphen in the template base,
    e.g. bases `foo_historical_bar` and `foo_rcp45_bar` will result in
    the template base `foo_historical-rcp45_bar`.

    If the template base contains a frequency keyword ('day') with
    surrounding underscores, the keyword will be replaced by `{frequency}`,
    otherwise `_{frequency}` will be added to the end of the template base.

    The time period of the template will always represent the whole period as
    covered by all files, given that the time periods can be parsed from the input
    file paths. A `computational_period` can be given as input argument to define a
    new time period. This period must be given in the format `yyyy-mm-dd/yyyy-mm-dd`
    and must be inside the original time period.

    The name of the index that is being computed will be added at the start of the
    template. If there is any period constrains it will be connected to the
    index with a hyphen, e.g, `{index}-jfm`.

    If these heuristic rules result in an unambiguous filename base, the template
    will be given as `{index}_{base}_{start}-{end}.nc`, or
    as `{index}_{base}.nc` if time period could not be determined.
    Otherwise the fallback `{index}_{frequency}.nc` will be returned.

    This works well with CMIP/CORDEX style filenames.

    Parameters
    ----------
    datafiles : list of string
        Input file names.
    computational_period : string
        A string with format `yyyy-mm-dd/yyyy-mm-dd` defining the time period.

    Returns
    -------
    output_template : string
        A format string suitable for use in the construction of an output
        filename.
    """
    frequency_keywords = ["day"]
    historical_keywords = ["hist", "historical"]
    scenario_keywords = [r"rcp\d{2}", r"ssp\d{3}"]

    def filename_stripper(path):
        # remove directory part...
        basename = os.path.basename(path)
        # ...and extension
        root, ext = os.path.splitext(basename)
        # split at _...
        parts = root.split("_")
        # and remove
        # first part (usually variable)
        # and last part (usually time)
        base = "_".join(parts[1:-1])
        try:
            time = [int(t) for t in parts[-1].split("-")]
            if len(time) == 1:
                time *= 2
        except ValueError:
            time = [None, None]
        return (base, time[0], time[1])

    def base_merger(bases):
        unique_bases = set(bases)
        if len(unique_bases) == 1:
            return unique_bases.pop()
        elif len(unique_bases) == 2:
            # Possible combination of historical and scenario data.
            hist_re = re.compile(rf"_({r'|'.join(historical_keywords)})_")
            scen_re = re.compile(rf"_({r'|'.join(scenario_keywords)})_")
            # Figure out which base is historical and which is scenario.
            hist_base = [b for b in unique_bases if hist_re.search(b)]
            scen_base = [b for b in unique_bases if scen_re.search(b)]
            if len(hist_base) == 1 and len(scen_base) == 1:
                hist_parts = hist_re.split(hist_base[0])
                scen_parts = scen_re.split(scen_base[0])
                if (
                    len(hist_parts) == 3
                    and len(scen_parts) == 3
                    and hist_parts[0] == scen_parts[0]
                    and hist_parts[2] == scen_parts[2]
                ):
                    concat_keywords = hist_parts[1] + "-" + scen_parts[1]
                    return f"{hist_parts[0]}_{concat_keywords}_{hist_parts[2]}"
        return None

    def add_frequency(base):
        # Replace occurrence of frequency keyword with {frequency}, or add _{frequency}.
        freq_re = re.compile(rf"_({'|'.join(frequency_keywords)})(?=_|$)")
        if freq_re.search(base):
            base = freq_re.sub("_{frequency}", base)
        else:
            base += "_{frequency}"
        return base

    def change_time_period(computational_period, start_bnds, end_bnds):
        time_parts = computational_period.split("/")
        time = [int(t.replace("-", "")) for t in time_parts]
        if (start_bnds <= time[0] <= time[1]) and time[1] <= end_bnds:
            return time[0], time[1]
        else:
            raise ValueError(
                f"Computational period <{computational_period}> must be on the "
                "form `yyyy-mm-dd/yyyy-mm-dd` and must be inside the time period "
                f"<{start_bnds}-{end_bnds}> given by the input files."
            )

    files = [filename_stripper(p) for p in datafiles]
    bases, starts, ends = zip(*files)
    base = base_merger(bases)
    if base:
        base_with_freq = add_frequency(base)
        if None in starts or None in ends:
            output_template = f"{{var_name}}_{base_with_freq}.nc"
        else:
            start = min(starts)
            end = max(ends)
            if computational_period is not None:
                start, end = change_time_period(computational_period, start, end)
            output_template = f"{{var_name}}_{base_with_freq}_{start}-{end}.nc"
    else:
        output_template = "{var_name}_{frequency}.nc"
    return output_template


def build_output_filename(index, datafiles, output_template, computational_period=None):
    if output_template is None:
        output_template = guess_output_template(datafiles, computational_period)
    period_specialization = index.period.specialization
    if period_specialization is not None:
        var_name = f"{{var_name}}-{period_specialization}"
    else:
        var_name = "{var_name}"
    drs_copy = index.metadata.output.drs.copy()
    drs_copy["var_name"] = var_name.format(**index.metadata.output.drs)
    drs_copy["frequency"] = index.period.label
    return output_template.format(**drs_copy)


def do_main(
    index_catalog,
    configuration,
    requested_indices,
    period,
    reference_period,
    computational_period,
    datafiles,
    output_template,
    sliced_mode,
    iterative_storage,
    scheduler,
):
    logging.debug("Preparing indices")
    indices = index_catalog.prepare_indices(
        requested_indices, period, reference_period, computational_period
    )
    for index_no, index in enumerate(indices):
        logging.info(
            "Starting calculations for index <"
            f"{requested_indices[index_no]}> in {index}"
        )
        logging.debug("Building output filename")
        output_filename = build_output_filename(
            index, datafiles, output_template, computational_period
        )
        setup_configuration(index, configuration)
        logging.debug("Preparing input data")
        input_data = prepare_input_data(datafiles, configuration)
        logging.debug("Calculating index")
        result = index(
            input_data,
            client=scheduler.client,
            sliced_mode=sliced_mode,
        )
        logging.info(f"Saving result in {os.path.abspath(output_filename)}")
        save(
            result,
            output_filename,
            iterative_storage,
            scheduler.client,
            conventions_override=(configuration is not None),
            configuration=configuration,
        )


def main():
    args = parse_args()
    setup_logging(args.log_level, args.verbose)

    logging.debug("Loading metadata")
    index_catalog, configuration = load_metadata(
        args.metadata_files, args.activate_config
    )
    if "list" in args.indices:
        print("Available indices are:")
        print(list(index_catalog.get_list()))
        return

    with setup_scheduler(args) as scheduler:
        logging.debug("Scheduler ready; starting main program.")
        start = time.time()
        try:
            do_main(
                index_catalog,
                configuration,
                args.indices,
                args.period,
                args.reference_period,
                args.computational_period,
                args.datafiles,
                args.output_template,
                args.sliced_mode,
                args.iterative_storage,
                scheduler,
            )
        finally:
            end = time.time()
            logging.info(f"Calculation took {end-start:.4f} seconds.")
        if args.keep_open:
            input("Press enter to close the cluster ")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from collections import namedtuple, OrderedDict

from cf_units import Unit
import iris


DENSITY_WATER = Unit("1000 kg m-3")


PrecipQty = namedtuple("PrecipQty", ["units", "is_rate_based", "is_mass_based"])


PRECIPITATION_INFO = OrderedDict(
    [
        ("lwe_precipitation_rate", PrecipQty(Unit("m s-1"), True, False)),
        ("lwe_thickness_of_precipitation_amount", PrecipQty(Unit("m"), False, False)),
        ("thickness_of_rainfall_amount", PrecipQty(Unit("m"), False, False)),
        ("precipitation_amount", PrecipQty(Unit("kg m-2"), False, True)),
        ("precipitation_flux", PrecipQty(Unit("kg m-2 s-1"), True, True)),
    ]
)


def is_precipitation(standard_name):
    return standard_name in PRECIPITATION_INFO.keys()


def get_standard_name_for_units(units):
    for standard_name, qty in PRECIPITATION_INFO.items():
        if units.is_convertible(qty.units):
            return standard_name
    raise ValueError(f"change_pr_units: Cannot handle new units {units}.")


def ensure_units_and_standard_name(units, standard_name):
    if isinstance(units, str):
        units = Unit(units)
    if units is None and standard_name is None:
        raise ValueError("No new units or new standard name is given.")
    if standard_name is None:
        standard_name = get_standard_name_for_units(units)
    try:
        units_dict = PRECIPITATION_INFO[standard_name]
    except KeyError:
        raise ValueError(f"Unknown standard_name {standard_name}.")
    if units is None:
        units = units_dict.units
    if not units.is_convertible(units_dict.units):
        raise ValueError(
            f"Cannot handle new units {units}. "
            f"Not fitting standard_name, {standard_name}."
        )
    return units, standard_name


def get_rate_conversion(old_qty, new_qty, integration_time):
    if old_qty.is_rate_based and not new_qty.is_rate_based:
        return integration_time
    if not old_qty.is_rate_based and new_qty.is_rate_based:
        return integration_time.invert()
    return 1.0


def get_mass_conversion(old_qty, new_qty, density):
    if old_qty.is_mass_based and not new_qty.is_mass_based:
        return density.invert()
    if not old_qty.is_mass_based and new_qty.is_mass_based:
        return density
    return 1.0


def change_pr_units(
    cube_or_coord,
    new_units=None,
    new_standard_name=None,
    integration_time=Unit("1 day"),
    density=DENSITY_WATER,
):
    """Converts precipitation units considering standard names.

    Converts precipitation units and data to requested units.
    Standard names are converted too, if required.

    Parameters
    ----------
    cube_or_coord: iris.cube.Cube or iris.coords.Coord
        precipitation data
    new_units: cf_units.Unit
        units to convert to
    new_standard_name: str
        standard name to convert to

    Returns
    -------
    iris.cube.Cube or iris.coords.Coord
       cube or coord with new units and standard name

    """
    new_units, new_standard_name = ensure_units_and_standard_name(
        new_units, new_standard_name
    )

    old_units, old_standard_name = ensure_units_and_standard_name(
        cube_or_coord.units, cube_or_coord.standard_name
    )

    old_qty = PRECIPITATION_INFO[old_standard_name]
    new_qty = PRECIPITATION_INFO[new_standard_name]

    conv = get_rate_conversion(
        old_qty, new_qty, integration_time
    ) * get_mass_conversion(old_qty, new_qty, density)

    conv_factor = (old_units * conv).convert(1.0, new_units)
    if isinstance(cube_or_coord, iris.cube.Cube):
        cube_or_coord.data = cube_or_coord.core_data() * float(conv_factor)
    elif isinstance(cube_or_coord, iris.coords.Coord):
        cube_or_coord.points *= float(conv_factor)
        if cube_or_coord.bounds is not None:
            cube_or_coord.bounds *= float(conv_factor)
    cube_or_coord.units = new_units
    cube_or_coord.standard_name = new_standard_name



#!/bin/env python

from jinja2 import Environment, PackageLoader, select_autoescape

from .metadata import load_metadata


class UnknownParameterKindException(Exception):
    pass


def raise_helper(msg):
    raise UnknownParameterKindException(msg)


def setup_templating():
    env = Environment(
        loader=PackageLoader("climix", "etc"),
        autoescape=select_autoescape(),
    )
    env.globals["raise"] = raise_helper
    return env


def main():
    catalog, config = load_metadata(None)
    distributions = {}
    for index in catalog.get_list():
        definition = catalog.get_index_definition(index)
        distribution = definition.distribution
        indices = distributions.setdefault(distribution, {})
        indices[index] = definition
    env = setup_templating()
    template = env.get_template("indices.md.jinja")
    print(template.render(distributions=distributions))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import numpy as np
import sparse


def dask_take_along_axis_chunk(x, idx, offset, x_size, axis):
    # Needed when idx is unsigned
    idx = idx.astype(np.int64)

    # Normalize negative indices
    idx = np.where(idx < 0, idx + x_size, idx)

    # A chunk of the offset dask Array is a numpy array with shape (1, ).
    # It indicates the index of the first element along axis of the current
    # chunk of x.
    idx = idx - offset

    # Drop elements of idx that do not fall inside the current chunk of x
    idx_filter = (idx >= 0) & (idx < x.shape[axis])
    idx[~idx_filter] = 0
    res = np.take_along_axis(x, idx, axis=axis)
    res[~idx_filter] = 0
    return sparse.COO(np.expand_dims(res, axis))


def dask_take_along_axis(x, index, axis):
    from dask.array.core import Array, blockwise, from_array, map_blocks

    if axis < 0:
        axis += x.ndim

    assert 0 <= axis < x.ndim

    assert (
        x.shape[:axis] + x.shape[axis + 1 :]
        == index.shape[:axis] + index.shape[axis + 1 :]
    )

    if np.isnan(x.chunks[axis]).any():
        raise NotImplementedError(
            "take_along_axis for an array with unknown chunks with "
            "a dask.array of ints is not supported"
        )

    # Calculate the offset at which each chunk starts along axis
    # e.g. chunks=(..., (5, 3, 4), ...) -> offset=[0, 5, 8]
    offset = np.roll(np.cumsum(x.chunks[axis]), 1)
    offset[0] = 0
    offset = from_array(offset, chunks=1)
    # Tamper with the declared chunks of offset to make blockwise align it with
    # x[axis]
    offset = Array(offset.dask, offset.name, (x.chunks[axis],), offset.dtype)

    # Define axis labels for blockwise
    x_axes = tuple(range(x.ndim))
    idx_label = (x.ndim,)  # arbitrary unused
    index_axes = x_axes[:axis] + idx_label + x_axes[axis + 1 :]
    offset_axes = (axis,)
    p_axes = x_axes[: axis + 1] + idx_label + x_axes[axis + 1 :]

    # Calculate the cartesian product of every chunk of x vs
    # every chunk of index
    p = blockwise(
        dask_take_along_axis_chunk,
        p_axes,
        x,
        x_axes,
        index,
        index_axes,
        offset,
        offset_axes,
        x_size=x.shape[axis],
        axis=axis,
        meta=sparse.COO(
            np.empty((0, 0), dtype=int),
            np.empty((), dtype=x.dtype),
            shape=(0,) * len(p_axes),
        ),
        dtype=x.dtype,
    )

    res = p.sum(axis=axis)

    res = map_blocks(lambda sparse_x: sparse_x.todense(), res, dtype=res.dtype)

    return res



# -*- coding: utf-8 -*-

from .index_functions import (  # noqa: F401
    CountLevelCrossings,
    CountOccurrences,
    CountJointOccurrencesPrecipitationTemperature,
    CountJointOccurrencesPrecipitationDoubleTemperature,
    CountJointOccurrencesTemperature,
    DiurnalTemperatureRange,
    ExtremeTemperatureRange,
    FirstOccurrence,
    InterdayDiurnalTemperatureRange,
    LastOccurrence,
    Percentile,
    Statistics,
    ThresholdedPercentile,
    ThresholdedStatistics,
    RunningStatistics,
    ThresholdedRunningStatistics,
    TemperatureSum,
)

from .percentile_functions import (  # noqa: F401
    CountPercentileOccurrences,
    CountThresholdedPercentileOccurrences,
)

from .spell_functions import (  # noqa: F401
    FirstSpell,
    SpellLength,
    SeasonStart,
    SeasonEnd,
    SeasonLength,
)



# -*- coding: utf-8 -*-

import logging
import time

import dask.array as da
from iris.analysis import Aggregator
import iris.coords

from gordias.dask_setup import progress
from .util import change_units


class PointLocalAggregator(Aggregator):
    def __init__(self, index_function, output_metadata, **kwargs):
        self.index_function = index_function
        self.output_metadata = output_metadata
        super().__init__(
            cell_method=None,
            call_func=self.index_function.call_func,
            lazy_func=self.index_function.lazy_func,
            **kwargs,
        )

    def update_metadata(self, cube, coords, **kwargs):
        super().update_metadata(cube, coords, **kwargs)
        cube.var_name = self.output_metadata.var_name
        cube.long_name = self.output_metadata.long_name
        cube.standard_name = self.index_function.standard_name
        cube.units = self.index_function.units
        cube.cell_methods = tuple(
            [
                iris.coords.CellMethod(cm.method, cm.name)
                for cm in self.output_metadata.cell_methods
            ]
        )

    def compute_pre_result(self, data, client, sliced_mode):
        if sliced_mode:
            logging.debug("Computing pre-result in sliced mode")
            stack = []
            end = time.time()
            cumulative = 0.0
            no_slices = len(data)
            for i, d in enumerate(data):
                result_id = f"{i+1}/{no_slices}"
                logging.info(f"Computing partial pre-result {result_id}")
                d = client.persist(d)
                progress(d)
                print()
                stack.append(d)
                start = end
                end = time.time()
                last = end - start
                cumulative += last
                eta = cumulative / (i + 1) * no_slices
                logging.info(
                    f"Finished {result_id} in (last cum eta): "
                    f"{last:4.0f} {cumulative:4.0f} {eta:4.0f}"
                )
            data = da.stack(stack, axis=0)
        else:
            logging.debug("Setting up pre-result in aggregate mode")
            start = time.time()
            data = client.persist(data)
            end = time.time()
            logging.debug(f"Setup completed in {end - start:4.0f}")
        return data

    def post_process(self, cube, data, coords, client, sliced_mode, **kwargs):
        data = self.compute_pre_result(data, client, sliced_mode)
        try:
            post_processor = self.index_function.post_process
        except AttributeError:
            # this index function does not require post processing
            pass
        else:
            cube, data = post_processor(cube, data, coords, **kwargs)
        cube = super().post_process(cube, data, coords, **kwargs)
        standard_name = self.output_metadata.standard_name
        unit_standard_name = standard_name
        proposed_standard_name = self.output_metadata.proposed_standard_name
        if unit_standard_name is None:
            unit_standard_name = proposed_standard_name
        change_units(cube, self.output_metadata.units, unit_standard_name)
        cube.standard_name = standard_name
        if proposed_standard_name is not None:
            cube.attributes["proposed_standard_name"] = proposed_standard_name
        for extra_coord in self.index_function.extra_coords:
            cube.add_aux_coord(extra_coord)
        return cube

    def pre_aggregate_shape(self, *args, **kwargs):
        return self.index_function.pre_aggregate_shape(*args, **kwargs)



import pytest
from contextlib import nullcontext as does_not_raise

from climix import main


TEST_FILE_LIST_TO_GUESS_OUTPUT_TEMPLATE = [
    (
        ["tas_foo_historical_bar_day_19700101.nc"],
        None,
        does_not_raise(
            "{var_name}_foo_historical_bar_{frequency}_19700101-19700101.nc"
        ),
    ),
    (
        ["tas_foo_historical_bar_day_19700101-19801231.nc"],
        None,
        does_not_raise(
            "{var_name}_foo_historical_bar_{frequency}_19700101-19801231.nc"
        ),
    ),
    (
        [
            "tas_foo_historical_bar_day_19700101-19801231.nc",
            "tas_foo_historical_bar_day_19800101-19901231.nc",
        ],
        None,
        does_not_raise(
            "{var_name}_foo_historical_bar_{frequency}_19700101-19901231.nc"
        ),
    ),
    (
        [
            "tas_foo_historical_bar_day_19900101-20001231.nc",
            "tas_foo_rcp26_bar_day_20010101-20101231.nc",
        ],
        None,
        does_not_raise(
            "{var_name}_foo_historical-rcp26_bar_{frequency}_19900101-20101231.nc"
        ),
    ),
    (
        [
            "tas_foo_historical_bar_day_19900101-20001231.nc",
            "tas_foo_rcp85_bar_day_20010101-20101231.nc",
        ],
        None,
        does_not_raise(
            "{var_name}_foo_historical-rcp85_bar_{frequency}_19900101-20101231.nc"
        ),
    ),
    (
        [
            "tas_foo_historical_bar_day_19900101-20001231.nc",
            "tas_foo_ssp119_bar_day_20010101-20101231.nc",
        ],
        None,
        does_not_raise(
            "{var_name}_foo_historical-ssp119_bar_{frequency}_19900101-20101231.nc"
        ),
    ),
    (
        [
            "tas_foo_historical_bar_day_19900101-20001231.nc",
            "tas_foo_ssp585_bar_day_20010101-20101231.nc",
        ],
        None,
        does_not_raise(
            "{var_name}_foo_historical-ssp585_bar_{frequency}_19900101-20101231.nc"
        ),
    ),
    (
        ["tas_foo_historical_bar_day.nc"],
        None,
        does_not_raise("{var_name}_foo_historical_bar_{frequency}.nc"),
    ),
    (
        ["tas_foo_historical_bar_day_notimerange.nc"],
        None,
        does_not_raise("{var_name}_foo_historical_bar_{frequency}.nc"),
    ),
    (
        ["tas_foo_historical_bar_notimerange.nc", "tas_foo_rcp45_bar_notimerange.nc"],
        None,
        does_not_raise("{var_name}_foo_historical-rcp45_bar_{frequency}.nc"),
    ),
    (["foobar.nc"], None, does_not_raise("{var_name}_{frequency}.nc")),
    (
        [
            "tas_foo_historical_bar_day_19700101-19801231.nc",
            "tas_foo_historical_bar_day_19800101-19901231.nc",
        ],
        "1970-02-03/1990-01-15",
        does_not_raise(
            "{var_name}_foo_historical_bar_{frequency}_19700203-19900115.nc"
        ),
    ),
    (
        ["tas_foo_historical_bar_day_19700101-19801231.nc"],
        "1970-02-03/1970-02-15",
        does_not_raise(
            "{var_name}_foo_historical_bar_{frequency}_19700203-19700215.nc"
        ),
    ),
    (
        [
            "tas_foo_historical_bar_day_19700101-19801231.nc",
            "tas_foo_historical_bar_day_19800101-19901231.nc",
        ],
        "1960-02-03/1990-01-15",
        pytest.raises(ValueError),
    ),
    (
        [
            "tas_foo_historical_bar_day_19700101-19801231.nc",
            "tas_foo_historical_bar_day_19800101-19901231.nc",
        ],
        "1970-02-03/2000-01-15",
        pytest.raises(ValueError),
    ),
    (
        [
            "tas_foo_historical_bar_day_19700101-19801231.nc",
        ],
        "1960-02-03/2000-01-15",
        pytest.raises(ValueError),
    ),
    (
        ["tas_foo_historical_bar_notimerange.nc", "tas_foo_rcp45_bar_notimerange.nc"],
        "1970-02-03/2000-01-15",
        does_not_raise("{var_name}_foo_historical-rcp45_bar_{frequency}.nc"),
    ),
]


@pytest.mark.parametrize(
    "file_list, computational_period, expected_output_template",
    TEST_FILE_LIST_TO_GUESS_OUTPUT_TEMPLATE,
)
def test_guess_output_template(
    file_list, computational_period, expected_output_template
):
    """Test ``guess_output_template``."""
    with expected_output_template:
        output_template = main.guess_output_template(file_list, computational_period)
        assert output_template == expected_output_template.enter_result



import dask.array as da
import iris
import numpy as np


def multicube_aggregated_by(cubes, coords, aggregator, **kwargs):
    # We assume all cubes have the same coordinates,
    # but a test needs to be added.
    groupby_coords = []
    dimension_to_groupby = None

    # We can't handle weights
    if isinstance(
        aggregator, iris.analysis.WeightedAggregator
    ) and aggregator.uses_weighting(**kwargs):
        raise ValueError(
            "Invalid Aggregation, multicube_aggregated_by() cannot use weights."
        )

    ref_cube = next(iter(cubes.values()))

    coords = ref_cube._as_list_of_coords(coords)
    for coord in sorted(coords, key=lambda coord: coord.metadata):
        if coord.ndim > 1:
            msg = (
                "Cannot aggregate_by coord %s as it is "
                "multidimensional." % coord.name()
            )
            raise iris.exceptions.CoordinateMultiDimError(msg)
        dimension = ref_cube.coord_dims(coord)
        if not dimension:
            msg = (
                'Cannot group-by the coordinate "%s", as its '
                "dimension does not describe any data." % coord.name()
            )
            raise iris.exceptions.CoordinateCollapseError(msg)
        if dimension_to_groupby is None:
            dimension_to_groupby = dimension[0]
        if dimension_to_groupby != dimension[0]:
            msg = "Cannot group-by coordinates over different dimensions."
            raise iris.exceptions.CoordinateCollapseError(msg)
        groupby_coords.append(coord)

    # Determine the other coordinates that share the same group-by
    # coordinate dimension.
    shared_coords = list(
        filter(
            lambda coord_: coord_ not in groupby_coords,
            ref_cube.coords(contains_dimension=dimension_to_groupby),
        )
    )

    # Determine which of each shared coord's dimensions will be aggregated.
    shared_coords_and_dims = [
        (coord_, index)
        for coord_ in shared_coords
        for (index, dim) in enumerate(ref_cube.coord_dims(coord_))
        if dim == dimension_to_groupby
    ]

    # Create the aggregation group-by instance.
    groupby = iris.analysis._Groupby(groupby_coords, shared_coords_and_dims)

    # Create the resulting aggregate-by cube and remove the original
    # coordinates that are going to be groupedby.
    # aggregateby_cube = iris.util._strip_metadata_from_dims(
    #     ref_cube, [dimension_to_groupby]
    # )
    key = [slice(None, None)] * ref_cube.ndim
    # Generate unique index tuple key to maintain monotonicity.
    key[dimension_to_groupby] = tuple(range(len(groupby)))
    key = tuple(key)
    # aggregateby_cube = aggregateby_cube[key]
    aggregateby_cube = ref_cube[key]
    for coord in groupby_coords + shared_coords:
        aggregateby_cube.remove_coord(coord)

    # Determine the group-by cube data shape.
    data_shape = list(ref_cube.shape + aggregator.pre_aggregate_shape(**kwargs))
    data_shape[dimension_to_groupby] = len(groupby)

    # Aggregate the group-by data.
    if aggregator.lazy_func is not None and ref_cube.has_lazy_data():

        def data_getter(cube):
            return cube.lazy_data()

        aggregate = aggregator.lazy_aggregate
        stack = da.stack
    else:

        def data_getter(cube):
            return cube.data

        aggregate = aggregator.aggregate
        stack = np.stack

    front_slice = (slice(None, None),) * dimension_to_groupby
    back_slice = (slice(None, None),) * (len(data_shape) - dimension_to_groupby - 1)

    if len(cubes) == 1:
        groupby_subcubes = map(
            lambda groupby_slice: ref_cube[front_slice + (groupby_slice,) + back_slice],
            groupby.group(),
        )

        def agg(cube):
            data = data_getter(cube)
            result = aggregate(data, axis=dimension_to_groupby, cube=cube, **kwargs)
            return result

    else:
        groupby_subcubes = map(
            lambda groupby_slice: {
                argname: cube[front_slice + (groupby_slice,) + back_slice]
                for argname, cube in cubes.items()
            },
            groupby.group(),
        )

        def agg(cubes):
            data = {argname: data_getter(cube) for argname, cube in cubes.items()}
            result = aggregate(data, axis=dimension_to_groupby, cube=cubes, **kwargs)
            return result

    result = list(map(agg, groupby_subcubes))
    aggregateby_data = stack(result, axis=dimension_to_groupby)

    # Add the aggregation meta data to the aggregate-by cube.
    aggregator.update_metadata(
        aggregateby_cube, groupby_coords, aggregate=True, **kwargs
    )
    # Replace the appropriate coordinates within the aggregate-by cube.
    (dim_coord,) = ref_cube.coords(
        dimensions=dimension_to_groupby, dim_coords=True
    ) or [None]
    for coord in groupby.coords:
        if (
            dim_coord is not None
            and dim_coord.metadata == coord.metadata
            and isinstance(coord, iris.coords.DimCoord)
        ):
            aggregateby_cube.add_dim_coord(coord.copy(), dimension_to_groupby)
        else:
            aggregateby_cube.add_aux_coord(coord.copy(), ref_cube.coord_dims(coord))

    # Attach the aggregate-by data into the aggregate-by cube.
    aggregateby_cube = aggregator.post_process(
        aggregateby_cube, aggregateby_data, coords, **kwargs
    )

    return aggregateby_cube
